#!/usr/bin/env python3
"""Codec wrapper for the emporda software (https://gici.uab.cat/GiciWebPage/downloads.php#emporda)
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2022/04/08"

import os
import shutil

import enb


class EmpordaAC(enb.icompression.LosslessCodec, enb.icompression.NearLosslessCodec,
                enb.icompression.JavaWrapperCodec, enb.icompression.GiciLibHelper):
    """Wrapper for the Empordà codec using a custom Arithmetic Coder

    (https://gici.uab.cat/GiciWebPage/downloads.php#emporda)
    """

    def __init__(self,
                 compressor_jar=os.path.join(os.path.dirname(__file__), "emporda_ac.jar"),
                 decompressor_jar=os.path.join(os.path.dirname(__file__), "emporda_ac.jar"),
                 qs=0, ec=1, cm=1, pm=0, wp=2048, up=2):
        """
        :param qs: sets the quantization step.
        :param ec: The encoder type that will encode the image.
            0.- Lossless without predictor + entropy encoder.
            1.- Lossless with predictor + entropy encoder.
            2.- Lossless and near-lossless with predictor predictor + entropy encoder.
            3.- Lossless and near-lossless with state-of-the-art predictor + entropy encoder.
        :param cm: context model
            0.- No context model is used.
            1.- Context modelling is used during the encoding process.
        :param pm: Probability model employed for the entropy coder.
            0.- The probability is estimated using a full division operation.
            1.- The probability is estimated using a division implemented through a quantized Look Up Table.
                This option must be used with -qlut option.
            2.- The probability is estimated using only bitwise operators and witout division.
                When this option is used -wp and -up parameters must be the same value of form 2^X.
        :param wp: Indicates the maximum number of symbols within the variable-size sliding windows
          that are employed for the Entropy Coder to compute the probability of the context.
          Must be of the form 2^X.
        :param up: Indicates the number of symbols coded before updating the context probability in the Entropy Coder.
            Must be of the form 2^X.
        """
        assert shutil.which("java") is not None, \
            f"The 'java' program was not found in the path, but is required by {self.__class__.__name__}. " \
            f"Please (re)install a JRE in the path and try again."
        super().__init__(compressor_jar=compressor_jar,
                         decompressor_jar=decompressor_jar,
                         param_dict=dict(qs=qs, ec=ec, cm=cm, pm=pm, wp=wp, up=up))

    def get_compression_params(self, original_path, compressed_path, original_file_info):
        assert original_file_info["bytes_per_sample"] == 2, \
            f"Only 16-bit samples are currently supported by {self.__class__.__name__}"
        assert not original_file_info["float"], \
            f"Only integer samples are currently supported by {self.__class__.__name__}"
        assert original_file_info["big_endian"], \
            f"Only big-endian samples are currently supported by {self.__class__.__name__}"

        return f"-Xmx256g -jar {self.compressor_jar} -c -i {original_path} -o {compressed_path} " \
               f"-ig {self.get_gici_geometry_str(original_file_info=original_file_info)} " \
               f"-qs {self.param_dict['qs']} " \
               f"-ec {self.param_dict['ec']} " \
               f"-cm {self.param_dict['cm']} " \
               f"-pm {self.param_dict['pm']} " \
               f"-wp {self.param_dict['wp']} " \
               f"-up {self.param_dict['up']}"

    def get_decompression_params(self, compressed_path, reconstructed_path, original_file_info):
        assert original_file_info["bytes_per_sample"] == 2, \
            f"Only 16-bit samples are currently supported by {self.__class__.__name__}"
        assert not original_file_info["float"], \
            f"Only integer samples are currently supported by {self.__class__.__name__}"
        assert original_file_info["big_endian"], \
            f"Only big-endian samples are currently supported by {self.__class__.__name__}"

        return f"-Xmx256g -jar {self.decompressor_jar} -d -i {compressed_path} -o {reconstructed_path} " \
               f"-ig {self.get_gici_geometry_str(original_file_info=original_file_info)} " \
               f"-qs {self.param_dict['qs']} " \
               f"-ec {self.param_dict['ec']} " \
               f"-cm {self.param_dict['cm']} " \
               f"-pm {self.param_dict['pm']} " \
               f"-wp {self.param_dict['wp']} " \
               f"-up {self.param_dict['up']}"

    @property
    def label(self):
        return "Emporda AC"


class EmpordaCCSDS(enb.icompression.LosslessCodec,
                   enb.icompression.JavaWrapperCodec,
                   enb.icompression.GiciLibHelper):
    """Wrapper for the Empordà codec using the CCSDS 123.0-B-1 entropy coders
    (block adaptive or sample adaptive), with the possibility of enabling and disabling
    the prediction stage.
    (https://gici.uab.cat/GiciWebPage/downloads.php#emporda)
    """
    SAMPLE_ADAPTIVE_ENCODER = 0
    BLOCK_ENCODER = 1
    OPTIONS_SAMPLE_ADAPTIVE_PATH = os.path.join(
        os.path.dirname(__file__), "options_sample_adaptive.txt")
    OPTIONS_BLOCK_ADAPTIVE_PATH = os.path.join(
        os.path.dirname(__file__), "options_block_adaptive.txt")

    def __init__(
            self,
            entropy_coder: int,
            skip_prediction: bool,
            compressor_jar: str = os.path.join(os.path.dirname(__file__), "emporda_ccsds.jar"),
            decompressor_jar: str = os.path.join(os.path.dirname(__file__), "emporda_ccsds.jar")):
        """
        :param entropy_coder: one of EmpordaNoPrediction.SAMPLE_ADAPTIVE_ENCODER or
          EmpordaNoPrediction.BLOCK_ENCODER, for the sample adaptive or the block adaptive
          entropy coder, respectively.
        :param skip_prediction: if True, the prediction stage is skipped
        :param compressor_jar, decompressor_jar: custom jar files implementing the emporda codec.
        """
        enb.icompression.JavaWrapperCodec.__init__(
            self,
            compressor_jar=compressor_jar,
            decompressor_jar=decompressor_jar,
            param_dict={"entropy_coder": entropy_coder,
                        "skip_prediction": 0 if not skip_prediction else 1})

    def get_compression_params(self, original_path, compressed_path, original_file_info):
        assert original_file_info["bytes_per_sample"] == 1, \
            f"Only 8-bit samples are currently supported by {self.__class__.__name__}"
        assert not original_file_info["float"], \
            f"Only integer samples are currently supported by {self.__class__.__name__}"
        assert original_file_info["big_endian"], \
            f"Only big-endian samples are currently supported by {self.__class__.__name__}"

        if self.param_dict["entropy_coder"] == self.SAMPLE_ADAPTIVE_ENCODER:
            options_file_path = self.OPTIONS_SAMPLE_ADAPTIVE_PATH
        elif self.param_dict["entropy_coder"] == self.BLOCK_ENCODER:
            options_file_path = self.OPTIONS_BLOCK_ADAPTIVE_PATH
        else:
            raise ValueError(f"Invalid entropy coder type {self.param_dict['entropy_coder']!r}")

        return (f"-Xmx256g -jar {self.decompressor_jar} "
                f"-c "
                f"-i {os.path.abspath(original_path)} "
                f"-ig {self.get_gici_geometry_str(original_file_info=original_file_info)} "
                f"-qm 0 "
                f"-o {os.path.abspath(compressed_path)} "
                f"-f {options_file_path} "
                f"-p {'0' if self.param_dict['skip_prediction'] else '1'}")

    def get_decompression_params(self, compressed_path, reconstructed_path, original_file_info):
        assert original_file_info["bytes_per_sample"] == 1, \
            f"Only 8-bit samples are currently supported by {self.__class__.__name__}"
        assert not original_file_info["float"], \
            f"Only integer samples are currently supported by {self.__class__.__name__}"
        assert original_file_info["big_endian"], \
            f"Only big-endian samples are currently supported by {self.__class__.__name__}"

        if self.param_dict["entropy_coder"] == self.SAMPLE_ADAPTIVE_ENCODER:
            options_file_path = self.OPTIONS_SAMPLE_ADAPTIVE_PATH
        elif self.param_dict["entropy_coder"] == self.BLOCK_ENCODER:
            options_file_path = self.OPTIONS_BLOCK_ADAPTIVE_PATH
        else:
            raise ValueError(f"Invalid entropy coder type {self.param_dict['entropy_coder']!r}")

        return (f"-Xmx256g -jar {self.decompressor_jar} "
                f"-d "
                f"-i {os.path.abspath(compressed_path)} "
                f"-ig {self.get_gici_geometry_str(original_file_info=original_file_info)} "
                f"-qm 0 "
                f"-o {os.path.abspath(reconstructed_path)} "
                f"-f {options_file_path} "
                f"-p {'0' if self.param_dict['skip_prediction'] else '1'}")

    @property
    def label(self):
        prediction_str = " NP" if self.param_dict["skip_prediction"] == 1 else ""
        ec_str = f"Sample adaptive" \
            if self.param_dict["entropy_coder"] == self.SAMPLE_ADAPTIVE_ENCODER \
            else "Block adaptive"
        return f"Emporda CCSDS {prediction_str} ({ec_str})"



#!/usr/bin/env python3
"""
Implementation of the classes for the enb.config.options and CLI interface

Option configuration in enb is centralized through enb.config.options. Several key aspects should be highlighted:

    - Properties defined in `enb.config.options` are used by enb modules, and can also be used by
      scripts using enb (host scripts).

    - Many core enb functions have optional arguments with default None values.
      Those functions will often substitute None for the corresponding value in enb.config.options,
      e.g., to locate the plot output directory.

    - Scripts using enb (host scripts) may alter values in enb.config.options, e.g., before calling enb methods.
      Properties are accessed and modified with `enb.config.options.property`
      and `enb.config.options.property = value`, respectively. You may want to use the `from enb.config import options` line
      in your host scripts to enable less verbosity.

    - The CLI can be used to set initial values of enb.config.options properties using `-*` and `--*` arguments.
      Running with `-h` any script that imports enb will show you detailed help on all available options and
      their default values.

    - The default values for enb.config.options and its CLI is obtained through enb.config.ini, described below.

An important note should be made about the interaction between enb.config.options and ray.
When ray spawns new (local or remote) processes to serve as workers, the Options singleton
is initialized for each of those process, with the catch that ray does **not** pass the user's CLI parameters.
Therefore, different enb.config.option values would be present in the parent script and the ray workers.
To mitigate this problem, the @`enb.parallel_ray.remote` decorator is provided in substitution of :meth:`ray.remote`
so that options at the time of calling the remote method are available to that method at
its regular location (enb.config.options).
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2019/08/04"

import os
import tempfile
import functools
import deprecation

from .. import default_base_dataset_dir
from .. import default_persistence_dir
from .. import default_output_plots_dir
from .. import default_analysis_dir
from .. import calling_script_dir
from . import ini
from . import singleton_cli as _singleton_cli

# Allowed message types
_logging_level_names = ["core", "error", "warning", "message", "verbose", "informative", "debug"]


class OptionsBase(_singleton_cli.SingletonCLI):
    """Global options for all modules, without any positional or required argument.
    """

    @property
    def non_default_properties(self):
        non_default_properties = dict()
        for k, v in self._name_to_property.items():
            try:
                if v == ini.get_key("enb.config.options", k):
                    continue
            except KeyError:
                pass
            non_default_properties[k] = v
        non_default_properties["verbose"] = self.verbose
        return non_default_properties

    def normalize_dir_value(self, value):
        if value[0] == os.sep:
            value = os.path.relpath(value, self.project_root)
        return value

    def __str__(self):
        s = "Summary of enb.config.options:\n\t- "
        s += "\n\t- ".join(f"{k:30s} = {repr(v)}"
                           for k, v in sorted(self.non_default_properties.items())
                           if k[0] != "_")
        return s


@_singleton_cli.property_class(OptionsBase)
class GeneralOptions:
    """Group of uncategorized options.
    """

    @OptionsBase.property("v", action="count")
    def verbose(self, value):
        """Be verbose? Repeat for more. Change at any time to increase the logger's verbosity.
        """
        from .. import log

        log.logger.selected_log_level = log.get_level(
            name=log.logger.level_message.name, lower_priority=float(value))

        return value

    @OptionsBase.property("ini", nargs="*", type=str, default=[])
    def extra_ini_paths(self, value):
        """Additional .ini files to be used to attain file-based configurations,
        in addition to the default ones (system, user and project).
        If defined more than once, the last definition sets the list instead of appending
        to a common list of extra ini paths.
        """
        pass


@_singleton_cli.property_class(OptionsBase)
class ExecutionOptions:
    """General execution options.
    """

    @OptionsBase.property(type=int)
    def cpu_limit(self, value):
        """Maximum number of CPUs to use for computation in this machine
        See https://miguelinux314.github.io/experiment-notebook/cluster_setup.html for
        details on how to set the resources employed in remote computation nodes.
        """
        if value is None:
            return value
        value = int(value)
        if value <= 0:
            value = None
        return value

    @OptionsBase.property("f", "overwrite", action="count")
    def force(self, value):
        """Force calculation of pre-existing results, if available?

        Note that should an error occur while re-computing a given index,
        that index is dropped from the persistent support.
        """
        return int(value)

    @OptionsBase.property("q", action="count")
    def quick(self, value):
        """Perform a quick test with a subset of the input samples?

        If specified q>0 times, a subset of the first q target indices is employed
        in most get_df methods from ATable instances
        """
        return int(value)

    @OptionsBase.property("render_only", action="store_true")
    def no_new_results(self, value):
        """If True, ATable's get_df method relies entirely on the loaded persistence data, no new rows are computed.
        This can be useful to speed up the rendering process, for instance to try different
        aesthetic plotting options. Use this option only if you know you need it.
        """
        return bool(value)

    @OptionsBase.property(type=int)
    def chunk_size(self, value):
        """Chunk size used when running ATable's get_df().
        Each processed chunk is made persistent before processing the next one.
        This parameter can be used to control the trade-off between error tolerance and overall speed.
        """
        return int(value)

    @OptionsBase.property(action=_singleton_cli.PositiveIntegerAction)
    def repetitions(self, value):
        """Number of repetitions when calculating execution times.

        This value allows computation of more reliable execution times in some experiments, but
        is normally most representative in combination with -s to use a single execution process at a time.
        """
        _singleton_cli.PositiveIntegerAction.assert_valid_value(value)
        return int(value)

    @OptionsBase.property(action="store_true")
    def report_wall_time(self, value):
        """If this flag is activated, the wall time instead of the CPU time is reported by default by
        tcall.get_status_output_time.
        """
        return bool(value)

    @OptionsBase.property(action="store_true")
    def force_sanity_checks(self, value):
        """If this flag is used, extra sanity checks are performed by enb during the execution of this script.
        The trade-off for rare error condition detection is a slower execution time.
        """
        return bool(value)

    @OptionsBase.property(nargs="+", type=str)
    def selected_columns(self, value):
        """List of selected column names for computation.

        If one or more column names are provided,
        all others are ignored. Multiple columns can be expressed,
        separated by spaces.
        """
        assert value, f"If provided, at least one column must be defined"

    @OptionsBase.property(type=float)
    def progress_report_period(self, value):
        """Default minimum time in seconds between progress report updates,
         when get_df() is invoked and computation is being processed in parallel.
         """
        return float(value)

    @OptionsBase.property(action="store_true")
    def disable_progress_bar(self, value):
        """If this flag is enabled, no progress bar is employed
        (useful to minimize the stdout volume of long-running experiments).
        """
        return bool(value)


@_singleton_cli.property_class(OptionsBase)
class DirOptions:
    """Options regarding default data directories.
    """

    @OptionsBase.property(action=_singleton_cli.ReadableDirAction, 
                          default=calling_script_dir)
    def project_root(self, value):
        """Project root path. It should not normally be modified.
        """
        _singleton_cli.ReadableDirAction.assert_valid_value(value)

    @OptionsBase.property(action=_singleton_cli.ReadableOrCreableDirAction, 
                          default=None)
    def base_dataset_dir(self, value):
        """Directory to be used as source of input files for indices in the get_df method
        of tables and experiments.

        It should be an existing, readable directory.
        """
        value = value or ini.get_key("enb.config.options", "base_dataset_dir")
        value = self.normalize_dir_value(value=value)
        _singleton_cli.ReadableOrCreableDirAction.assert_valid_value(value)
        return value

    @OptionsBase.property(action=_singleton_cli.WritableOrCreableDirAction,
                          default=None)
    def persistence_dir(self, value):
        """Directory where persistence files are to be stored.
        """
        value = value or ini.get_key("enb.config.options", "persistence_dir")
        value = self.normalize_dir_value(value=value)
        _singleton_cli.WritableOrCreableDirAction.assert_valid_value(value)
        return value

    # Reconstructed version dir
    @OptionsBase.property(action=_singleton_cli.WritableOrCreableDirAction)
    def reconstructed_dir(self, value):
        """Base directory where a copy of the reconstructed versions of data are to be stored.
        """
        value = value or ini.get_key("enb.config.options", "reconstructed_dir")
        value = self.normalize_dir_value(value=value)
        _singleton_cli.WritableOrCreableDirAction.assert_valid_value(value)
        return value
    
    @OptionsBase.property(action=_singleton_cli.WritableOrCreableDirAction)
    def compressed_copy_dir(self, value):
        """Base directory where a copy of the compressed versions of data are to be stored.
        """
        value = value or ini.get_key("enb.config.options", "compressed_copy_dir")
        value = self.normalize_dir_value(value=value)
        _singleton_cli.WritableOrCreableDirAction.assert_valid_value(value)
        return value

    # Versioned data dir
    @OptionsBase.property(action=_singleton_cli.WritableOrCreableDirAction,
                          default=None)
    def base_version_dataset_dir(self, value):
        """Base dir for versioned folders.
        """
        value = value or ini.get_key("enb.config.options", "version_dataset_dir")
        value = self.normalize_dir_value(value=value)
        _singleton_cli.WritableOrCreableDirAction.assert_valid_value(value)
        return value

    # Temp dir
    for default_tmp_dir in ["/dev/shm", "/var/run", tempfile.gettempdir()]:
        try:
            _singleton_cli.WritableDirAction.assert_valid_value(default_tmp_dir)
            break
        except AssertionError:
            pass
    else:
        default_tmp_dir = os.path.expanduser("~/enb_tmp")

    @OptionsBase.property(action=_singleton_cli.WritableOrCreableDirAction,
                          default=default_tmp_dir)
    def base_tmp_dir(self, value):
        """Temporary dir used for intermediate data storage.

        This can be useful when experiments make heavy use of tmp and memory is limited,
        avoiding out-of-RAM crashes at the cost of potentially slower execution time.

        The dir is created when defined if necessary.
        """
        _singleton_cli.WritableOrCreableDirAction.assert_valid_value(value)

    # Base dir for external binaries (e.g., codecs or other tools)
    default_external_binary_dir = os.path.join(calling_script_dir, "bin")
    default_external_binary_dir = default_external_binary_dir \
        if _singleton_cli.ReadableDirAction.check_valid_value(default_external_binary_dir) else None

    @OptionsBase.property(action=_singleton_cli.ReadableDirAction, default=default_external_binary_dir,
                          required=False)
    def external_bin_base_dir(self, value):
        """External binary base dir.

        In case a centralized repository is defined at the project or system level.
        """
        _singleton_cli.ReadableDirAction.assert_valid_value(value)

    # Output plots dir
    @OptionsBase.property(
        action=_singleton_cli.WritableOrCreableDirAction,
        default=default_output_plots_dir)
    def plot_dir(self, value):
        """Directory to store produced plots.
        """
        value = self.normalize_dir_value(value=value)
        _singleton_cli.WritableOrCreableDirAction.assert_valid_value(value)
        return value

    # Output analysis dir
    @OptionsBase.property(action=_singleton_cli.WritableOrCreableDirAction, default=default_analysis_dir)
    def analysis_dir(self, value):
        """Directory to store analysis results.
        """
        value = self.normalize_dir_value(value=value)
        _singleton_cli.WritableOrCreableDirAction.assert_valid_value(value)
        return value


@_singleton_cli.property_class(OptionsBase)
class RayOptions:
    """Options related to the ray library, used for parallel/distributed computing
    only when --ssh_cluster_csv_path (or, equivalently --ssh_csv) are employed.
    """

    @OptionsBase.property("ssh_csv", type=str, default=None)
    def ssh_cluster_csv_path(self, value):
        """Path to the CSV file containing a enb ssh cluster configuration.
        See https://miguelinux314.github.io/experiment-notebook/installation.html.
        """
        if not os.path.exists(value):
            print(f"Selected ssh_cluster_csv_path={repr(value)}, but it is not a valid file. "
                  f"Setting to None instead.")
            value = None
        return value

    @OptionsBase.property(action="store_true")
    def disable_swap(self, value):
        """If this flag is used, then swap memory will not be allowed by ray. By default, swap memory is enabled.
        Note that your system may become unstable if swap memory is used (specially a big portion thereof).
        """
        return bool(value)

    @OptionsBase.property(type=str)
    def worker_script_name(self, value):
        """Base name of ray's worker scripts, invoked to run tasks in parallel processes.
        You don't need to change this unless you want to use custom ray workers.
        """
        if value != os.path.basename(value):
            raise SyntaxError("The worker_script_name parameter must be a base name, i.e., a file name "
                              f"including any extension, and without any path indication. Found {value} instead")
        return str(value)

    @OptionsBase.property(type=str)
    def preshutdown_wait_seconds(self, value):
        """A wait period can be held before shutting down ray. This allows displaying messages produced by
        child processes (e.g., stack traces) in case of abrupt termination of enb client code.
        """
        return float(value) if float(value) <= 0 else 0

    @OptionsBase.property(action=_singleton_cli.PositiveIntegerAction)
    def ray_port(self, value):
        """Ray port and first port that need to be open in case a cluster
        is to be set up. Refer to https://miguelinux314.github.io/experiment-notebook/installation.html
        for further information on this.
        """
        _singleton_cli.PositiveIntegerAction.assert_valid_value(value)
        return int(value)

    @OptionsBase.property(action=_singleton_cli.PositiveIntegerAction)
    def ray_port_count(self, value):
        """Total number of consecutive ports that can be assumed to be open after `ray_port`.
        For instance, if `ray_port` is 11000 and `ray_port_count` is 1000, then
        ports 11000-11999 will be used for parallelization and (if so-configured) enb clusters.
        """
        _singleton_cli.PositiveIntegerAction.assert_valid_value(value)
        return int(value)

    @OptionsBase.property(action="store_true")
    def no_remote_mount_needed(self, value):
        """If this flag is used, the calling script's project root path is assumed to
        be valid AND synchronized (e.g., via NFS). By default, remote mounting via sshfs and vde2 is employed.
        """
        return bool(value)


@_singleton_cli.property_class(OptionsBase)
class LoggingOptions(OptionsBase):
    """Options controlling what and how is printed and/or logged to files.
    """

    @OptionsBase.property(type=str, choices=_logging_level_names)
    def selected_log_level(self, value):
        """Maximum log level / minimum priority required when printing messages.
        """
        return str(value)

    @OptionsBase.property(type=str, choices=_logging_level_names)
    def default_print_level(self, value):
        """Selects the default log level equivalent to a regular print-like message. It is most effective
        when combined with log_print set to True.
        """
        return str(value)

    @OptionsBase.property(type=bool, choices=[True, False])
    def log_level_prefix(self, value):
        """If True, logged messages include a prefix, e.g., based on their priority.
        """
        from .. import log

        log.logger.show_prefixes = bool(value)
        return log.logger.show_prefixes

    @OptionsBase.property(type=str, choices=_logging_level_names)
    def show_prefix_level(self, value):
        """If True, a prefix indicating the priority level of each message is shown 
        preceding those messages.
        """
        from .. import log

        log.logger.show_prefix_level = log.logger.get_level(value)
        
    @OptionsBase.property(type=bool, choices=[True, False])
    def force_live_progress(self, value):
        """If True, the progress panels shown by ATable are forced to be updated live
        even if the output is not an interactive terminal.
         
        Enabling this option will significantly increase the size of any log file
        to which the output is redirected, because the output will include the periodic
        screen writes and deletions produced, e.g., every second.
        """
        return bool(value)


class Options(GeneralOptions, ExecutionOptions, DirOptions, RayOptions, LoggingOptions):
    """Class of the `enb.config.options` object, which exposes
    options for all modules, allowing CLI-based parameter setting.

    Classes wishing to expand the set of global options can be defined above,
    using the @OptionsBase.property decorator for new properties.
    Making :class:`Options` inherit from those classes is optional,
    but allows IDEs to automatically
    detect available properties in `enb.config.options`.

    Parameters in this class should defined so that no positional or otherwise mandatory
    arguments. This is due to interactions with ray for parallelization purposes, which
    results in `sys.argv` differing in the orchestrating and host processes.
    """
    pass


@deprecation.deprecated(deprecated_in="0.2.7", removed_in="0.3.1")
def get_options(from_main=False):
    """Deprecated - use `from enb.config import options`.
    """
    return Options()


def set_options(new_option_dict):
    """Update global options with a dictionary of values
    """
    if Options() is not new_option_dict:
        for k, v in new_option_dict.__dict__.items():
            Options().__setattr__(k, v)


def propagates_options(f):
    """Decorator for local (as opposed to ray.remote) functions so that they
    propagate options properly to child workers.
    The decorated function must accept an "options" argument.
    Furthermore, the current working dir is set to the project root so that
    any relative paths stored are correctly handled.
    """

    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        os.chdir(kwargs["options"].project_root)
        set_options(kwargs["options"])
        return f(*args, **kwargs)

    return wrapper



#!/usr/bin/env python3
"""Experiment notebook (enb) library.

Please see https://github.com/miguelinux314/experiment-notebook
for further information.
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2020/03/31"

# pylint: disable wrong-import-position

import warnings as _warnings
import atexit as _atexit
import os as _os
import sys as _sys
import appdirs as _appdirs
import numpy as _np
import builtins as _builtins

# Make all warnings errors
_np.seterr(all="raise")
_warnings.simplefilter('error', UserWarning)

# Current installation dir of enb
enb_installation_dir = _os.path.dirname(_os.path.abspath(__file__))

# User configuration dir (e.g., ~/.config/enb in many linux distributions)
user_config_dir = _os.path.join(
    _os.path.abspath(_os.path.expanduser(_appdirs.user_config_dir())), "enb")
if not _os.path.exists(user_config_dir):
    _os.makedirs(user_config_dir, exist_ok=True)
if not _os.path.isdir(user_config_dir):
    print(
        f"[enb.__init__.py]: user_config_dir={user_config_dir} "
        f"is not a directory or could not be created.")

# Absolute, real path to the calling script's dir.
# Configuration files present here will overwrite
# those in `user_config_dir`.
calling_script_dir = _os.path.realpath(
    _os.path.dirname(_os.path.abspath(_sys.argv[0]))) \
    if _sys.argv[0] else _os.getcwd()

# Are we currently running the main enb CLI or the CLI for a host script?
# True means main enb CLI.
is_enb_cli = _os.path.basename(_sys.argv[0]) in ["__main__.py", "enb"]

# Data dir
default_base_dataset_dir = _os.path.join(calling_script_dir, "datasets")
# Persistence dir (default: persistence/)
default_persistence_dir = _os.path.join(calling_script_dir, "persistence")
# Plots dir (default: plots/<script_name>)
default_output_plots_dir = _os.path.join(
    calling_script_dir if not is_enb_cli else ".",
    "plots",
    _os.path.basename(_sys.argv[0])[:-len(".py") if _sys.argv[0].endswith(".py") else None])
# Analysis dir (default: analysis/<script_name>)
default_analysis_dir = _os.path.join(
    calling_script_dir if not is_enb_cli else ".",
    "analysis",
    _os.path.basename(_sys.argv[0])[:-len(".py") if _sys.argv[0].endswith(".py") else None])

# Fix getcwd for the specific case enb is imported from the sphinx documentation tool
if _os.path.basename(_sys.argv[0]) == "sphinx-build":
    default_persistence_dir = _os.path.join(_os.getcwd(), "build",
                                            f"persistence_{_os.path.basename(_sys.argv[0])}")
    default_base_dataset_dir = _os.path.join(_os.getcwd(), "build", "datasets")
    default_output_plots_dir = _os.path.join(_os.getcwd(), "build", "plots")
    default_analysis_dir = _os.path.join(_os.getcwd(), "build", "analysis")

# pylint: disable=wrong-import-position
# Basic tools among core modules
from . import misc
# Global configuration modules
from . import config
# Logging tools
from . import log
from .log import logger
# Paralellization modules
from . import parallel
from . import parallel_ray
# Live progress display
from . import progress

# Temporary fix until the dill library is fixed
parallel.parallel_fix_dill_crash()

# Setup logging so that it is used from here on. Done here to avoid circular dependencies.
logger.selected_log_level = log.get_level(
    name=config.options.selected_log_level,
    lower_priority=config.options.verbose)
logger.show_prefixes = config.options.log_level_prefix
logger.show_prefix_level = logger.get_level(
    name=config.options.show_prefix_level)

# Redirect print calls to enb's logger
_builtins.print = logger.print_to_log

# Remaining core modules
## Keystone ATable features
from . import atable
## Basic Experiment features
from . import sets
from . import experiment
from . import isets
## Data analysis (e.g., plotting) modules
from . import plotdata
from . import render
from . import aanalysis
## Compression modules
from . import compression
# Imports for backwards compatibility
from .compression import icompression
from .compression import fits
from .compression import png
from .compression import jpg
from .compression import pgm
from .compression import tarlite

# Plugin and template support
from . import plugins

# For backwards compatibility
isets.FITSVersionTable = fits.FITSVersionTable
isets.PNGCurationTable = png.PNGCurationTable
isets.raw_path_to_png = png.raw_path_to_png
isets.render_array_png = png.render_array_png

# Kept for backwards compatibility
icompression.FITSVersionTable = fits.FITSVersionTable
icompression.FitsWrapperCodec = fits.FITSWrapperCodec
icompression.PNGWrapperCodec = png.PNGWrapperCodec
icompression.PGMWrapperCodec = pgm.PGMWrapperCodec

# Setup to be run only when enb is imported in the main process
if not parallel_ray.is_parallel_process():
    log.show_banner()

    if config.ini.all_ini_paths:
        log.info(f"Additional .ini files employed: {', '.join(repr(p) for p in config.ini.all_ini_paths)}.")

    __file__ = _os.path.abspath(__file__)
    if not is_enb_cli:
        _os.chdir(calling_script_dir)

    if parallel_ray.is_ray_enabled():
        logger.info(f"Using ray for parallelization (CPU limit: {config.options.cpu_limit})\n")
        _atexit.register(parallel_ray.stop_ray)
        # The list of modules loaded so far passed to any possible ray remote
        # nodes so that they don't attempt to load them again.
        # pylint: disable=protected-access
        config.options._initial_module_names = list(
            m.__name__ for m in _sys.modules.values() if
            hasattr(m, "__name__"))
    else:
        logger.info(f"Using pathos for parallelization (CPU limit: {config.options.cpu_limit})\n")

    misc.capture_usr1()

# Run the setter functions on the default values too, allowing validation and normalization
config.options.update(config.options, trigger_events=True)

# Remote process invoked with ray need to have some imports added
parallel_ray.fix_imports()



import enb.plugins


class MontsecPlugin(enb.plugins.plugin.PluginJava):
    name = "montsec"
    label = "Wrapper for the montsec codec"
    tags = {"data compression", "image", "codec"}
    authors = ["Pau Quintas Torra", "Xavier Fernandez Mellado"]
    contrib_authors = ["Group on Interactive Coding of Images (GICI)"]
    tested_on = {"linux", "macos"}



#!/usr/bin/env python3
"""Wrapper for Nayuki's arithmetic coder, originally downloaded from
https://github.com/nayuki/Reference-arithmetic-coding/
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2021/06/01"

import os
from enb import icompression


class ArithmeticCodec(icompression.WrapperCodec, icompression.LosslessCodec):
    """Wrapper for Nayuki's arithmetic coder, originally downloaded from
    https://github.com/nayuki/Reference-arithmetic-coding/
    """

    def __init__(self, compressor_path=None, decompressor_path=None):
        if compressor_path is None:
            compressor_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "ArithmeticCompress")
        if decompressor_path is None:
            decompressor_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "ArithmeticDecompress")
        super().__init__(compressor_path=compressor_path, decompressor_path=decompressor_path)

    def get_compression_params(self, original_path, compressed_path, original_file_info):
        try:
            os.remove(compressed_path)
        except FileNotFoundError:
            pass
        return f"{original_path} {compressed_path}"

    def get_decompression_params(self, compressed_path, reconstructed_path, original_file_info):
        try:
            os.remove(reconstructed_path)
        except FileNotFoundError:
            pass
        return f"{compressed_path} {reconstructed_path}"

    @property
    def label(self):
        return "Arithmetic Coder"



import os
import stat
import glob
import enb.plugins


class CompressionExperimentTemplate(enb.plugins.Template):
    """Template for lossless data compression experiments.
    """
    name = "lossless-compression"
    author = ["Miguel Hernández-Cabronero"]
    label = "Template for lossless compression experiments"
    tags = {"project", "image", "documentation", "data compression"}
    tested_on = {"linux", "windows"}

    @classmethod
    def install(cls, installation_dir, overwrite_destination=False, fields=None):
        # Save
        super().install(installation_dir=installation_dir,
                        overwrite_destination=overwrite_destination,
                        fields=fields)
        for py_path in glob.glob(os.path.join(installation_dir, "*.py")):
            os.chmod(py_path, stat.S_IREAD | stat.S_IWRITE | stat.S_IEXEC)

        # Copy the default enb.ini file into the project folder
        enb.plugins.install("enb.ini", installation_dir, overwrite=True, automatic_import=False)



import os
import enb.plugins

class BWTPlugin(enb.plugins.PluginMake):
    name = "bwt"
    label = "Application of the Burrows-Wheeler Transform (BWT)"
    tags = {"data compression", "codec"}
    contrib_authors = ["Michael Dipperstein"]
    contrib_reference_urls = ["https://github.com/MichaelDipperstein/bwt"]
    tested_on = {"linux"}

    @classmethod
    def build(cls, installation_dir):
        super().build(installation_dir=installation_dir)
        os.rename(os.path.join(installation_dir, "sample"),
                  os.path.join(installation_dir, "bwt"))



import sys
import os
import shutil
import stat
import glob
import subprocess
import enb.plugins
from enb.config import options


class EntropyCodecComparisonTemplate(enb.plugins.Template):
    """Template for a comparison of (lossless) entropy codecs.
    """
    name = "entropy-codec-comparison"
    author = ["Miguel Hernández-Cabronero"]
    label = "Template for the comparison of entropy codecs"
    tags = {"project", "data compression"}
    tested_on = {"linux"}

    additional_plugins = ["fse", "zip", "lz4", "arithmetic-coder", "huffman"]

    @classmethod
    def install(cls, installation_dir, overwrite_destination=False, fields=None):
        super().install(installation_dir=installation_dir,
                        overwrite_destination=overwrite_destination,
                        fields=fields)
        for py_path in glob.glob(os.path.join(installation_dir, "*.py")):
            os.chmod(py_path, stat.S_IREAD | stat.S_IWRITE | stat.S_IEXEC)

        # Copy the default enb.ini file into the project folder
        target_enb_ini = os.path.join(installation_dir, "enb.ini")
        if os.path.exists(target_enb_ini) and not options.force:
            enb.logger.warn(f"File {target_enb_ini} already exists. Not overwriting.")
        else:
            shutil.copyfile(os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))),
                                         f"config", "enb.ini"),
                            os.path.join(installation_dir, f"enb.ini"))
            
        for p in cls.additional_plugins:
            with enb.logger.message_context(f"Installing plugin {repr(p)}"):
                invocation = f"{sys.executable} -m enb plugin install {p} {os.path.join(installation_dir, 'plugins', p)}"
                status, output = subprocess.getstatusoutput(invocation)
                if status != 0:
                    raise Exception(f"Status = {status} != 0.\n"
                                    f"Input=[{invocation}].\n"
                                    f"Output=[{output}]")



#!/usr/bin/env python3
"""JPEG manipulation (e.g., curation) tools.
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2023/11/09"

from enb.compression.png import PNGCurationTable

class JPEGCurationTable(PNGCurationTable):
    """Given a directory tree containing JPEG images, copy those images into
    a new directory tree in raw BSQ format adding geometry information tags to
    the output names recognized by `enb.isets.load_array_bsq`.
    """
    dataset_files_extension = "jpg"



#!/usr/bin/env python3
"""Unit tests for the enb's ray subsystem.
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2021/08/18"

import unittest
import enb
import time

class TestRay(unittest.TestCase):
    if enb.parallel_ray.is_ray_enabled():
        def test_ray_is_automatically_started(self):
            try:
                import ray
            except ImportError:
                print("ray not installed: skipping")
                
            enb.parallel_ray.init_ray()
            assert ray.is_initialized()

        def test_progress_run(self):
            try:
                import ray
            except ImportError:
                print("ray not installed: skipping")
                
            self.test_ray_is_automatically_started()

            sleep_time = 0.5

            @ray.remote
            def wait_a(_):
                time.sleep(sleep_time)

            @ray.remote
            def wait_b(_):
                time.sleep(sleep_time / 2)

            target_ids = [wait_a.remote(ray.put(_)) for _ in range(5)] \
                         + [wait_b.remote(ray.put(_)) for _ in range(5)]

            pg = enb.parallel.ProgressiveGetter(id_list=target_ids, iteration_period=0.6)

            for i, _ in enumerate(pg):
                assert len(pg.completed_ids) < len(target_ids)
                assert i < 10, pg

            # All ids should be immediately ready after iterating the ProgressiveGet instance
            ray.get(target_ids, timeout=0)
    else:
        def test_skip(self):
            enb.logger.verbose("Skipping ray test because a ray ssh cluster is not selected.")


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
"""Codec wrapper for the FPC lossless image coder
"""
__author__ = "Òscar Maireles and Miguel Hernández-Cabronero"
__since__ = "2021/08/01"

import os
import enb


class Fpc(enb.icompression.LosslessCodec, enb.icompression.NearLosslessCodec, enb.icompression.WrapperCodec):
    """Wrapper for the FPC codec
    All data types integer and float 16, 32, 64 can be compressed 
    """

    def __init__(self, fpc_binary=os.path.join(os.path.dirname(__file__), "fpc"), compression_level=25):
        super().__init__(compressor_path=fpc_binary,
                         decompressor_path=fpc_binary,
                         param_dict=dict(compression_level=compression_level))
        """
        :param compression_level: 1-29, being 29 the maximum data reduction       
        """

    @property
    def label(self):
        return "FPC"

    def get_compression_params(self, original_path, compressed_path, original_file_info):
        return f" {self.param_dict['compression_level']} <{original_path}>  {compressed_path}"

    def get_decompression_params(self, compressed_path, reconstructed_path, original_file_info):
        return f" < {compressed_path} > {reconstructed_path} "



import os
import datetime

import enb
from enb.config import options


class EnbIniTemplate(enb.plugins.Template):
    """Copy a full enb.ini configuration in the destination project folder.
    """
    name = "enb.ini"
    author = ["Miguel Hernández-Cabronero"]
    label = "Copy a full enb.ini configuration in the destination project folder."
    tags = {"project"}
    tested_on = {"linux", "windows"}

    @classmethod
    def build(cls, installation_dir):
        full_ini_path = os.path.join(enb.enb_installation_dir, "config", "enb.ini")
        output_path = os.path.join(installation_dir, "enb.ini")
        if os.path.exists(output_path) and not options.force:
            raise ValueError(f"Output file {output_path} exists. Refusing to overwrite.")

        with open(full_ini_path, "r") as input_file, open(output_path, "w") as output_file:
            output_file.write(
                f"# Project created {datetime.datetime.now()}.\n"
                f"# Default option values can be changed by uncommenting them below.\n"
                f"# NOTE: CLI options overwrite the ones defined in this file.\n"
                "\n")
            output_file.write("".join(f"# {line}" for line in input_file.readlines()))



#!/usr/bin/env python3
"""Codec wrapper for the Zstandard lossless image coder
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2021/07/12"

import os
import enb
from enb import tarlite
import tempfile
import subprocess

class Zstandard(enb.icompression.LosslessCodec, enb.icompression.NearLosslessCodec, enb.fits.FITSWrapperCodec):
    """Wrapper for the Zstandard codec
    All data types integer and float 16, 32, 64 can be compressed
    """

    def __init__(self, compression_level=19, zstd_binary=os.path.join(os.path.dirname(__file__), "zstd")):
        """
        :param compression_level: 1-22, being 22 the maximum data reduction
        """
        assert 1 <= int(compression_level) <= 22, f"Invalid compression level {compression_level}"
        super().__init__(compressor_path=zstd_binary,
                         decompressor_path=zstd_binary,
                         param_dict=dict(compression_level=compression_level))

    @property
    def label(self):
        return f"Zstandard {self.param_dict['compression_level']}"

    def assert_valid_data_type(self, original_file_info):
        """Verify that the input data type is supported, else raise an exception
        """
        assert not original_file_info["float"], f"Only integer samples are supported by {self.__class__.__name__}"
        assert original_file_info["big_endian"], f"Only big-endian samples are supported by {self.__class__.__name__}"

    def get_compression_params(self, original_path, compressed_path, original_file_info):
        self.assert_valid_data_type(original_file_info=original_file_info)
        return f"-{self.param_dict['compression_level']} " \
               f"{'--ultra ' if self.param_dict['compression_level'] > 19 else ''}" \
               f"-f {original_path}  -o {compressed_path}"

    def get_decompression_params(self, compressed_path, reconstructed_path, original_file_info):
        return f"-d  -f {compressed_path} -o {reconstructed_path}"


class Zstandard_train(Zstandard):
    """Wrapper for the Zstandard codec, which produces a dictionary file for the input data and then employs
    it for compression. The generated dictionary is included in the compressed data size measurements.

    All data types integer and float 16, 32, 64 can be compressed
    """

    def assert_valid_data_type(self, original_file_info):
        """Verify that the input data type is supported, else raise an exception
        """
        super().assert_valid_data_type(original_file_info=original_file_info)
        assert original_file_info["bytes_per_sample"] == 4, \
            f"Only 32-bit samples are supported by {self.__class__.__name__}"

    def compress(self, original_path, compressed_path, original_file_info):
        self.assert_valid_data_type(original_file_info=original_file_info)

        with tempfile.TemporaryDirectory(dir=enb.config.options.base_tmp_dir) as tmp_dir:
            dict_path = os.path.join(tmp_dir, "dict.zstd")
            dict_compressed_path = os.path.join(tmp_dir, "compressed.zstd")

            # Generate dictionary
            invocation = f"{self.compressor_path} --train {original_path} --optimize-cover -o {dict_path}"
            status, output = subprocess.getstatusoutput(invocation)
            if status != 0:
                raise Exception(f"Status = {status} != 0.\n"
                                f"Input=[{invocation}].\n"
                                f"Output=[{output}]")

            # Compress using that dictionary
            invocation = f" {self.compressor_path} " \
                         f"{'--ultra ' if int(self.param_dict['compression_level']) > 19 else ''}" \
                         f"-{self.param_dict['compression_level']} -D {dict_path} -f {original_path} " \
                         f"-o {dict_compressed_path}"
            status, output = subprocess.getstatusoutput(invocation)
            if status != 0:
                raise Exception(f"Status = {status} != 0.\n"
                                f"Input=[{invocation}].\n"
                                f"Output=[{output}]")

            # Unite the compressed data and the side information
            tarlite.tarlite_files(input_paths=(dict_path, dict_compressed_path), output_tarlite_path=compressed_path)

    def decompress(self, compressed_path, reconstructed_path, original_file_info):
        with tempfile.TemporaryDirectory(dir=enb.config.options.base_tmp_dir) as tmp_dir:
            dict_path = os.path.join(tmp_dir, "dict.zstd")
            dict_compressed_path = os.path.join(tmp_dir, "compressed.zstd")

            # Separate the compressed data from the side information
            tarlite.untarlite_files(input_tarlite_path=compressed_path, output_dir_path=tmp_dir)

            # Decompress
            invocation = f"{self.compressor_path} -d -D {dict_path} -f {dict_compressed_path} -o {reconstructed_path}"
            status, output = subprocess.getstatusoutput(invocation)
            if status != 0:
                raise Exception(f"Status = {status} != 0.\n"
                                f"Input=[{invocation}].\n"
                                f"Output=[{output}]")

    @property
    def label(self):
        return f"Zstandard pretrain {self.param_dict['compression_level']}"



import enb.plugins


class BitshufflePlugin(enb.plugins.Plugin):
    name = "bitshuffle"
    label = "Codec wrapper that applies bitshuffle before another codec"
    tags = {"data compression", "codec"}
    contrib_authors = ["Andrew Collete et al."]
    contrib_reference_urls = ["https://github.com/h5py/h5py"]
    required_pip_modules = ["bitshuffle"]
    tested_on = {"linux"}



#!/usr/bin/env python3
"""
Wrappers for the Kakadu codec.
For instructions on downloading and installing visit:
https://github.com/miguelinux314/experiment-notebook/blob/master/enb/plugins/plugin_kakadu/README.md
"""
__author__ = "Natalia Blasco, Ester Jara, Artemis Llabrés and Miguel Hernández-Cabronero"
__since__ = "2021/08/01"

import math
import numpy as np
import os
import tempfile
import enb
from enb import icompression
from enb import isets


class Kakadu2D(icompression.WrapperCodec, icompression.LosslessCodec, icompression.LossyCodec):
    """Kakadu JPEG 2000 wrapper that applies the 5/3 IWT or the 9/7 DWT (spatially) when lossless is True or False,
    respectively. No spectral decorrelation is applied, unless the apply_ycc flag is selected.

    Note that subclasses may overwrite the get_atk_params method to configure specific spatial transforms.
    """
    # When searching for exact PSNR values, this error is tolerated
    psnr_tolerance = 1e-3
    max_search_iterations = 32

    def __init__(self, ht=False, spatial_dwt_levels=5,
                 lossless=None, bit_rate=None, quality_factor=None, psnr=None,
                 apply_ycc=False, num_threads=0):
        """Kakadu wrapper that does not apply decorrelation across bands, even if more than one is present.

            :param ht: if True, the high-throughput version of the Kakadu codec is used
            :param spatial_dwt_levels: number of spatial discrete wavelet transform levels.
              Must be between 0 and 33.
            :param lossless: if True, the compression will be lossless, and bit_rate, quality_factor or
              psnr can not be
              used. If None, a lossless compression will be performed, unless
              bit_rate/quality_factor/psnr are set,
              in which case lossless will be set to False and the compression will be lossy.
            :param bit_rate: target bits per sample for each component of the image. If bit_rate is set,
              then lossless must be None or False.
            :param quality_factor: target quality factor for a lossy compression.
              If quality_factor is set then lossless
              must be None or False. Must be between 0 and 100.
            :param psnr: target peak signal-to-noise ratio for a lossy compression.
              A binary search is performed to find which quality factor will result in the
              indicated PSNR with the tolerance given by self.psnr_tolerance.
              If psnr is set then lossless must be None or False.
            :param num_threads: number of threads with which kakadu is to be run. 0 means single thread.
        """
        assert isinstance(ht, bool), "HT must be a boolean (True/False)"
        assert spatial_dwt_levels in range(0, 34), \
            f"Invalud number of spatial DWT levels {spatial_dwt_levels}"

        lossless = lossless if lossless is not None else (
                bit_rate is None and quality_factor is None and psnr is None)
        if lossless:
            assert all(v is None for v in [bit_rate, quality_factor, psnr]), \
                f"Cannot set bitrate, quality factor or PSNR when lossless is requested."
        else:
            assert sum(1 for v in (bit_rate, quality_factor, psnr) if v is not None) <= 1, \
                f"Only one of bitrate, quality factor and PSNR can be set at a time."
            if bit_rate is not None:
                assert bit_rate > 0, f"Invalid bit rate {bit_rate}"
            if quality_factor is not None:
                assert 0 < quality_factor <= 100, f"Invalid quality factor {quality_factor}"
            if psnr is not None:
                assert psnr > 0, f"Invalid psnr {psnr}"
            lossless = False

        assert num_threads == int(num_threads), f"Invalid number of threads {num_threads}"
        assert num_threads >= 0, f"Invalid number of threads {num_threads}"

        param_dict = dict(
            ht=ht,
            spatial_dwt_levels=spatial_dwt_levels,
            lossless=lossless,
            bit_rate=bit_rate,
            quality_factor=quality_factor,
            psnr=psnr,
            num_threads=num_threads)
        if apply_ycc:
            param_dict["apply_ycc"] = True
        icompression.WrapperCodec.__init__(
            self,
            compressor_path=os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                         "kdu_compress"),
            decompressor_path=os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                           "kdu_expand"),
            param_dict=param_dict)

    def get_atk_params(self, original_file_info):
        """Get the ATK parameters that allow customization of the spatial 2D DWT applied to the data.
        If this method returns an empty string, then the default type of DWT is applied
        (e.g., the reversible 5/3 IWT for lossless compression, or the irreversible 9/7 CDF for lossy compression.
        Note that the Creversible=yes or Creversible=no is already provided by self.get_compression_params()
        """
        return ""

    def get_compression_params(self, original_path, compressed_path, original_file_info):
        if self.param_dict["num_threads"] > 1 and not enb.config.options.report_wall_time:
            enb.logger.warn("Careful! Measuring kakadu's process CPU time with multiple threads. "
                            "Add `options.report_wall_time = True` to your script or "
                            "invoke it with --report_wall_time to enable wall clock time measurements.")
        if original_file_info["float"]:
            raise ValueError("This enb codec does not support floating point data. "
                             "Please refer to the kakadu documentation "
                             "and perform manual compression")

        precision_bits = original_file_info['bytes_per_sample'] * 8

        return f"-i {original_path}*{original_file_info['component_count']}" \
               f"@{original_file_info['width'] * original_file_info['height'] * original_file_info['bytes_per_sample']} " \
               f"-o {compressed_path} -no_info -full -no_weights " \
               f"-num_threads {self.param_dict['num_threads']} " \
               f"Corder=LRCP " \
               f"Clevels={self.param_dict['spatial_dwt_levels']} " \
               f"Clayers={original_file_info['component_count']} " \
               f"Creversible={'yes' if self.param_dict['lossless'] else 'no'} " \
               f"{self.get_atk_params(original_file_info=original_file_info)} " \
               f"Cycc={'yes' if 'apply_ycc' in self.param_dict and self.param_dict['apply_ycc'] else 'no'} " \
               f"Sdims=\\{{{original_file_info['height']},{original_file_info['width']}\\}} " \
               f"Nprecision={precision_bits} " \
               f"Sprecision={precision_bits} " \
               f"Nsigned={'yes' if original_file_info['signed'] else 'no'} " \
               f"Ssigned={'yes' if original_file_info['signed'] else 'no'} " \
            + (f"Qstep=0.000000001 " if self.param_dict['bit_rate'] is not None else "") \
            + f"{'Cmodes=HT' if self.param_dict['ht'] else ''} " \
              f"{'-rate ' + str(self.param_dict['bit_rate'] * original_file_info['component_count']) if self.param_dict['bit_rate'] else ''}" \
              f"{'Qfactor=' + str(self.param_dict['quality_factor']) if self.param_dict['quality_factor'] else ''}"

    def get_decompression_params(self, compressed_path, reconstructed_path, original_file_info):
        return f"-i {compressed_path} -o {reconstructed_path} -raw_components " \
               f"-num_threads {self.param_dict['num_threads']}"

    def compress(self, original_path, compressed_path, original_file_info=None):
        if self.param_dict['psnr'] is not None:
            with tempfile.NamedTemporaryFile() as tmp_file:
                br_a = 0
                br_b = 32
                iteration = 0

                psnr_error = float("inf")
                while psnr_error > self.psnr_tolerance and iteration <= self.max_search_iterations:
                    iteration += 1
                    self.param_dict['bit_rate'] = (br_b + br_a) / 2
                    compression_results = icompression.WrapperCodec.compress(
                        self, original_path, compressed_path, original_file_info=original_file_info)
                    self.decompress(
                        compressed_path,
                        reconstructed_path=tmp_file.name,
                        original_file_info=original_file_info)

                    max_error = (2 ** (8 * original_file_info["bytes_per_sample"])) - 1
                    dtype = isets.iproperties_row_to_numpy_dtype(original_file_info)
                    original_array = np.fromfile(original_path, dtype=dtype).astype(np.int64)
                    reconstructed_array = np.fromfile(tmp_file.name, dtype=dtype).astype(np.int64)
                    actual_bps = 8 * os.path.getsize(compressed_path) / original_file_info[
                        "samples"]

                    mse = np.average(((original_array - reconstructed_array) ** 2))
                    psnr = 10 * math.log10((max_error ** 2) / mse) if mse > 0 else float("inf")
                    if self.param_dict['psnr'] > psnr:
                        br_a = min(self.param_dict['bit_rate'], actual_bps)
                    else:
                        br_b = max(self.param_dict['bit_rate'], actual_bps)

                    psnr_error = abs(self.param_dict['psnr'] - psnr)

        else:
            compression_results = icompression.WrapperCodec.compress(
                self, original_path, compressed_path, original_file_info=original_file_info)

        return compression_results

    def decompress(self, compressed_path, reconstructed_path, original_file_info=None):
        tmp_file_list = []
        for i in range(0, original_file_info['component_count']):
            tmp_file_list.append(tempfile.NamedTemporaryFile(
                dir=enb.config.options.base_tmp_dir, suffix=".raw", delete=True))
        output_path_str = ",".join(tmp_file.name for tmp_file in tmp_file_list)

        decompression_results = icompression.WrapperCodec.decompress(
            self, compressed_path, reconstructed_path=output_path_str,
            original_file_info=original_file_info)

        with open(reconstructed_path, "wb") as output_file:
            for tmp_file in tmp_file_list:
                with open(tmp_file.name, "rb") as component_file:
                    output_file.write(component_file.read())
        decompression_results.reconstructed_path = reconstructed_path

        for tmp_file in tmp_file_list:
            tmp_file.close()

        return decompression_results

    @property
    def label(self):
        rate_str = "" if not self.param_dict["bit_rate"] else f" R = {self.param_dict['bit_rate']}"
        psnr_str = "" if not self.param_dict["psnr"] else f" PSNR {self.param_dict['psnr']}"
        ycc_str = "" if "apply_ycc" not in self.param_dict or not self.param_dict[
            "apply_ycc"] else " YCC"
        return f"Kakadu {'HT' if self.param_dict['ht'] else ''}" \
               f"{'lossless' if self.param_dict['lossless'] else 'lossy'}{rate_str}{psnr_str}{ycc_str}"


class Kakadu2DHaar(Kakadu2D, icompression.LosslessCodec):
    """Kakadu JPEG 2000 wrapper that applies the spatial reversible or irreversible
    Haar transform when lossless is True or False,
    respectively. No spectral decorrelation is applied, unless the apply_ycc flag is selected.
    """

    def get_atk_params(self, original_file_info):
        return f"Catk=3 Kkernels:I3={'R' if self.param_dict['lossless'] else 'I'}2X2"

    @property
    def label(self):
        return f"{super().label} Haar"


class Kakadu2D97M(Kakadu2D, icompression.LosslessCodec):
    """Kakadu JPEG 2000 wrapper that applies the reversible 9/7-M DWT in the spatial 2D transform.
    """

    def get_atk_params(self, original_file_info):
        """Get the ATK filter for the reversible 9/7-M.
        """
        if not self.param_dict["lossless"]:
            enb.logger.warn(f"Warning: using {self.__class__.__name__} with lossy mode")
        return " Catk=7 " \
               " Ksteps:I7=\\{4,-1,4,7\\},\\{2,-1,1,1\\} " \
               " Kcoeffs:I7=0.0625,-0.5625,-0.5625,0.0625,0.25,0.25 " \
               " Kreversible:I7=yes " \
               " Ksymmetric:I7=yes "

    @property
    def label(self):
        return f"{super().label} 9/7-M"


class KakaduMCT(Kakadu2D):
    """Kakadu JPEG 2000 wrapper that applies the reversible 5/3 IWT or the irreversible 9/7 DWT
    (both spatially and spectrally) when lossless is True or False, respectively.
    """

    def __init__(self, ht=False, spatial_dwt_levels=5, spectral_dwt_levels=5, lossless=None,
                 bit_rate=None, quality_factor=None, psnr=None, mct_offset=0,
                 num_threads=0):
        """
        :param spectral_dwt_levels: number of spectral discrete wavelet transform levels. Must be between 0 and 32.
        :param mct_offset: integer offset subtracted from the input samples before compression.
          For unsigned data, the average pixel value or half the maximum pixel value are often good choices.
          For signed data, 0 or the average pixel value are often good choices.
        """
        assert 0 <= spectral_dwt_levels <= 32, f"Invalid number of spectral levels"
        assert mct_offset == int(mct_offset)
        Kakadu2D.__init__(self, ht=ht, spatial_dwt_levels=spatial_dwt_levels, lossless=lossless,
                          bit_rate=bit_rate, quality_factor=quality_factor, psnr=psnr,
                          num_threads=num_threads)
        self.param_dict["spectral_dwt_levels"] = spectral_dwt_levels
        self.param_dict["mct_offset"] = mct_offset

    def get_compression_params(self, original_path, compressed_path, original_file_info):
        return Kakadu2D.get_compression_params(
            self,
            original_path=original_path,
            compressed_path=compressed_path,
            original_file_info=original_file_info) + \
            f" Mcomponents={original_file_info['component_count']} " \
            f"Mstage_inputs:I1=\\{{0,{original_file_info['component_count'] - 1}\\}} " \
            f"Mstage_outputs:I1=\\{{0,{original_file_info['component_count'] - 1}\\}} " \
            f"Mstage_collections:I1=\\{{{original_file_info['component_count']},{original_file_info['component_count']}\\}} " \
            f"Mstage_xforms:I1=\\{{DWT," \
            + ('1' if self.param_dict['lossless'] else '0') + \
            f",4,0,{self.param_dict['spectral_dwt_levels']}\\}} " \
            f"Mvector_size:I4={original_file_info['component_count']} " \
            f"Mvector_coeffs:I4={self.param_dict['mct_offset']} Mnum_stages=1 Mstages=1"

    def get_decompression_params(self, compressed_path, reconstructed_path, original_file_info):
        return f"-i {compressed_path} -o {reconstructed_path} -raw_components "

    @property
    def label(self):
        return super().label.replace("Kakadu", "Kakadu MCT")



#!/usr/bin/env python3
"""Tools to display live progress of ATable instances based on the rich library.
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2024/01/01"

import os
import math
import inspect
import collections
import rich
import rich.progress
import rich.panel
import rich.layout
import rich.console
import rich.progress_bar

import enb.experiment
from .config import options
from .config.aini import managed_attributes


def is_progress_enabled():
    """Return True if and only if all conditions for displaying progress are met:
    - verbose being level 1 (verbose) or 2 (info)
    - disable_progress_bar is set to False
    """
    return options.verbose in (1, 2) and not options.disable_progress_bar


@managed_attributes
class ProgressTracker(rich.live.Live):
    """Keep track of the progress of an ATable's (incl. Experiments') get_df.
    """
    # Style for the panel border surrounding each progress track
    style_border = "#adadad bold"

    # Style for the completed portion of the progress bar while it's being filled
    style_bar_complete = "#9b5ccb bold"
    # Style for the incomplete portion of the progress bar
    style_bar_incomplete = "#252525"
    # Style for the bar once the task is finished
    style_bar_finished = "#f3ac05"

    # Style for the panel title when the instance being tracked is an enb.sets.FilePropertiesTable subclass
    style_title_dataset = "#45e193"
    # Style for the panel title when the instance being tracked is an enb.experiment.Experiment subclass
    style_title_experiment = "#13bf00"
    # Style for the panel title when the instance being tracked is an enb.aanalysis.Analyzer subclass
    style_title_analyzer = "#1990ff"
    # Style for the panel title when the instance being tracked is an enb.aanalysis.AnalyzerSummary subclass
    style_title_summary = "#23cfff"
    # Style for the panel title when the instance being tracked is any other type of enb.atable.ATable instance
    style_title_atable = "#9b59ff"
    # Style for the panel title when the instance being tracked is not an enb.atable.ATable subclass
    style_title_other = "#cdabff"

    # Style for the text labels (e.g., "Rows", "Chunks")
    style_text_label = "#787878"
    # Style for the text separators (e.g., ":", "/")
    style_text_separator = "#505050"
    # Style for the text indicating units or scale (e.g., "s", "%")
    style_text_unit = "#707070"
    # Style for the text displaying the number of completed elements
    style_text_completed = "#bcbcbc bold"
    # Style for the text displaying the total number of elements
    style_text_total = "#bcbcbc"
    # Style for the text displaying speed
    style_text_speed = "#bcbcbc"
    # Style for the text displaying completion percentage
    style_text_percentage = "#bcbcbc"
    # Style for text displaying time values (not units)
    style_text_time = "#bcbcbc"

    # Style for the spinner
    style_spinner = "#9b5ccb bold"

    # Keep references to the current ProgressTracker instances in a LIFO
    _current_instance_stack = collections.deque()

    def __init__(self, atable, row_count: int, chunk_size: int):
        """
        :param atable: ATable subclass instance for which the progress is to be tracked
        :param row_count: total number of rows that need to be computed
        :param chunk_size: chunk size (any non-positive number
          is also interpreted as a chunk size equal to row_count)
        """
        self.atable = atable
        self.row_count = row_count
        self.chunk_size = min(row_count, chunk_size if chunk_size > 0 else row_count)

        row_progress_column = _RowProgressColumn(self)
        elapsed_and_expected_column = _ElapsedAndExpectedColumn(self)
        chunk_progress_column = _ChunkProgressColumn(self)
        self.row_progress_bar = _RowProgressBar(self)

        # Define the upper and lower progress rows
        self.upper_progress = rich.progress.Progress(
            rich.progress.SpinnerColumn(style=self.style_spinner),
            row_progress_column,
            _SeparatorColumn(self, width=1),
            elapsed_and_expected_column,
        )
        self.lower_progress = rich.progress.Progress(
            _SeparatorColumn(self, width=1),
            chunk_progress_column,
            _SeparatorColumn(self, width=1),
            self.row_progress_bar,
        )

        # Create the tasks that keep track of the completed rows and chunks
        self.row_task_id = self.upper_progress.add_task("Rows", total=self.row_count, pulse=False)
        self.row_task = [t for t in self.upper_progress.tasks if t.id == self.row_task_id][0]
        self.chunk_task_id = self.lower_progress.add_task(f"Chunk", total=math.ceil(
            self.row_count / self.chunk_size) if self.chunk_size else 0)
        self.chunk_task = [t for t in self.lower_progress.tasks if t.id == self.chunk_task_id][0]

        # Fix some column widths for a visually pleasant distribution
        first_column_width = max(
            len(rich.text.Text.from_markup(row_progress_column.get_render_str()).plain),
            len(rich.text.Text.from_markup(chunk_progress_column.get_render_str()).plain))
        row_progress_column.width = first_column_width
        chunk_progress_column.width = first_column_width

        # Display the progress rows one above the other
        self.group = rich.console.Group(self.upper_progress, self.lower_progress)

        title = (f"[{self._instance_to_title_style(atable)}]"
                 f"{atable.__class__.__name__}"
                 f"[/{self._instance_to_title_style(atable)}]")

        # Display the caller information for info and higher verbose level
        if enb.logger.level_active(enb.logger.level_info.name):
            for record in inspect.stack():
                if os.path.dirname(record.filename).startswith(os.path.dirname(__file__)):
                    continue
                title += (f"[not bold][{enb.logger.style_info}]"
                          f" < {os.path.basename(record.filename)}:{record.lineno}"
                          f"[/{enb.logger.style_info}][/not bold]")
                break

        # Combine all widgets into a single renderable for this progress tracker
        self.panel = rich.panel.Panel(
            self.group,
            title=title,
            title_align="left",
            expand=True,
            border_style=self.style_border)

        # Force live progress report if requested via CLI or .ini files
        super().__init__(
            self.panel,
            console=rich.console.Console(
                force_terminal=True,
                force_interactive=True)
            if enb.config.options.force_live_progress else None)

        # Keep track of current instances so that the console of the most recent progress can be employed
        ProgressTracker._current_instance_stack.append(self)

    def complete_chunk(self):
        """Add 1 to the number of completed chunks if a chunk task has been defined.
        """
        if self.chunk_task_id is not None:
            self.lower_progress.advance(self.chunk_task_id)

    def update_chunk_completed_rows(self, chunk_completed_rows):
        """Set the number of rows completed for the current chunk.
        """
        previously_completed = self.chunk_task.completed * self.chunk_size \
            if self.chunk_task is not None else 0
        self.upper_progress.update(
            self.row_task_id,
            completed=previously_completed + chunk_completed_rows)
        self.row_progress_bar.total = self.row_count
        self.row_progress_bar.completed = self.row_task.completed

    @property
    def chunk_count(self):
        """Get the number of chunks defined for this progress tracking stage.
        """
        return math.ceil(self.row_count / self.chunk_size) if self.chunk_size else 0

    @classmethod
    @property
    def console(cls):
        """Return the console instance for the current instance (the only live instance) of ProgressTracker,
        or None if none are available.
        """
        try:
            return ProgressTracker._current_instance_stack[-1].console
        except IndexError:
            return None

    def _instance_to_title_style(self, instance):
        """Return the current configured title style for the
        type of the instance whose progress is being tracked.
        """
        if isinstance(instance, enb.experiment.Experiment):
            return self.style_title_experiment
        if isinstance(instance, enb.sets.FilePropertiesTable):
            return self.style_title_dataset
        if isinstance(instance, enb.aanalysis.Analyzer):
            return self.style_title_analyzer
        if isinstance(instance, enb.aanalysis.AnalyzerSummary):
            return self.style_title_summary
        if isinstance(instance, enb.atable.ATable):
            return self.style_title_atable
        return self.style_title_other

    def __del__(self):
        last = ProgressTracker._current_instance_stack.pop()
        if last is not self:
            enb.logger.warn(
                "Warning! The ProgressTracker instance stack seems not not be in the right order.")


class _ProgressColumn:
    """Base class for progress columns that can consider both the row and the chunk progress.
    """

    def __init__(self, progress_tracker: ProgressTracker):
        super().__init__()
        self.progress_tracker = progress_tracker

    @property
    def chunk_task(self):
        return self.progress_tracker.chunk_task

    @property
    def row_task(self):
        return self.progress_tracker.row_task

    @property
    def chunk_size(self):
        return self.progress_tracker.chunk_size

    @property
    def chunk_count(self):
        return self.progress_tracker.chunk_count


class _ProgressTextColumn(_ProgressColumn, rich.progress.TextColumn):
    """Base class for progress columns that display text.
    """

    def __init__(self, progress_tracker: ProgressTracker, width: int = 0):
        rich.progress.TextColumn.__init__(self, "")
        self.progress_tracker = progress_tracker
        self.width = width

    def render(self, task):
        return rich.progress.Text.from_markup(f"{{:{self.width}s}}".format(self.get_render_str()))

    def get_render_str(self) -> str:
        raise NotImplemented

    def seconds_to_time_str(self, seconds: float, decimals=1, markup=True) -> str:
        """Convert the passed number of seconds to an HH:MM:SS format.
        :param seconds: the number of seconds (can be fractional)
        :param decimals: the number of decimals to which the number of seconds
          is rounded to.
        :param markup: if True, the values and units are surrounded by style markup tags.
        """
        seconds = round(seconds, decimals)
        second_fraction = seconds - int(seconds)
        minutes, seconds = divmod(int(seconds), 60)
        hours, minutes = divmod(minutes, 60)

        if markup is True:
            value_tag_start = f"[{self.progress_tracker.style_text_time}]"
            value_tag_end = f"[/{self.progress_tracker.style_text_time}]"
            unit_tag_start = f"[{self.progress_tracker.style_text_unit}]"
            unit_tag_end = f"[/{self.progress_tracker.style_text_unit}]"
        else:
            value_tag_start = ""
            value_tag_end = ""
            unit_tag_start = ""
            unit_tag_end = ""

        second_fraction_formatter = f"{{:.{decimals}f}}"

        return ((f"{value_tag_start}{hours:02d}{value_tag_end}{unit_tag_start}h{unit_tag_end} "
                 if hours else "")
                + (f"{value_tag_start}{minutes:02d}{value_tag_end}{unit_tag_start}m{unit_tag_end} "
                   if minutes else "")
                + (f"{value_tag_start}{seconds:02d}{value_tag_end}{unit_tag_start}s{unit_tag_end}"
                   if minutes or hours else
                   (f"{value_tag_start}{seconds}{value_tag_end}"
                    + (f"{unit_tag_start}s{unit_tag_end}" if decimals <= 0 else "")
                    + ((f"{value_tag_start}."
                        f"{second_fraction_formatter.format(second_fraction)[2:]}"
                        f"{value_tag_end}"
                        f"{unit_tag_start}s{unit_tag_end}")
                       if decimals > 0 else "")
                    ))
                )


class _RowProgressColumn(_ProgressTextColumn):
    """Display the row completion progress and the speed, if available.
    """

    def get_render_str(self) -> str:
        # Add the completed and total number of rows
        total_row_digits = len(str(self.row_task.total))
        row_count_formatter = f"{{:0{total_row_digits}d}}"
        speed_formatter = "{:>11s}"
        render_str = (f"[{self.progress_tracker.style_text_label}]"
                      f"Rows[/{self.progress_tracker.style_text_label}]"
                      f"  [{self.progress_tracker.style_text_separator}]"
                      f":[/{self.progress_tracker.style_text_separator}] "
                      f"[{self.progress_tracker.style_text_completed}]"
                      f"{row_count_formatter.format(self.row_task.completed)}"
                      f"[/{self.progress_tracker.style_text_completed}]"
                      f"[{self.progress_tracker.style_text_separator}]"
                      f"/[/{self.progress_tracker.style_text_separator}]"
                      f"[{self.progress_tracker.style_text_total}]"
                      f"{self.row_task.total}[/{self.progress_tracker.style_text_total}]"
                      f" ")

        # Add extra spaces so that any row count up to 5 digits generates the same width
        render_str += " " * 2 * (5 - total_row_digits)

        # Add speed if available
        if self.row_task.speed:
            # Padding needs to take into account the markup to maintain alignment
            formatted_length = (
                    11
                    + (2 * len(self.progress_tracker.style_text_unit) + len("[][/]")
                       if not self.row_task.finished else 0)
                    + 2 * len(self.progress_tracker.style_text_speed) + len("[][/]")
                    + 2 * len(self.progress_tracker.style_text_unit) + len("[][/]")
            )
            speed_formatter = f"{{:>{formatted_length}s}}"
            render_str += speed_formatter.format(
                (f"[{self.progress_tracker.style_text_unit}]"
                 f"+[/{self.progress_tracker.style_text_unit}]"
                 if not self.row_task.finished else ' ') +
                f"[{self.progress_tracker.style_text_speed}]"
                f"{self.row_task.speed:.2f}[/{self.progress_tracker.style_text_speed}]"
                f"[{self.progress_tracker.style_text_unit}]"
                f"/s[/{self.progress_tracker.style_text_unit}]")
        else:
            speed_formatter = "{:>11s}"
            render_str += speed_formatter.format("")

        return render_str


class _ElapsedAndExpectedColumn(_ProgressTextColumn):
    """Display the total elapsed and expected times.
    """

    def get_render_str(self) -> str:
        render_str = (f"[{self.progress_tracker.style_text_total}]"
                      f"{self.seconds_to_time_str(self.chunk_task.elapsed, markup=True)}"
                      f"[/{self.progress_tracker.style_text_total}]")
        if self.row_task.time_remaining:
            render_str += (f" [{self.progress_tracker.style_text_separator}]"
                           f"+~[/{self.progress_tracker.style_text_separator}] "
                           f"[{self.progress_tracker.style_text_total}]"
                           f"{self.seconds_to_time_str(self.row_task.time_remaining, markup=True)}"
                           f"[/{self.progress_tracker.style_text_total}]")
        return render_str


class _ChunkProgressColumn(_ProgressTextColumn):
    """Displays the chunk completion, size and row completion percentage.
    """

    def get_render_str(self) -> str:
        # Calculate the completed/total chunk string
        self.chunk_count_format = f"{{:0{len(str(self.chunk_count))}d}}"
        render_str = (f"[{self.progress_tracker.style_text_label}]Chunks"
                      f"[/{self.progress_tracker.style_text_label}]"
                      f"[{self.progress_tracker.style_text_separator}]"
                      f":[/{self.progress_tracker.style_text_separator}] "
                      f"[{self.progress_tracker.style_text_completed}]"
                      f"{self.chunk_count_format.format(self.chunk_task.completed)}"
                      f"[/{self.progress_tracker.style_text_completed}]"
                      f"[{self.progress_tracker.style_text_separator}]"
                      f"/[/{self.progress_tracker.style_text_separator}]"
                      f"[{self.progress_tracker.style_text_total}]"
                      f"{self.chunk_count_format.format(self.chunk_task.total)}"
                      f"[/{self.progress_tracker.style_text_total}]")

        # Add extra spaces so that any row count up to 5 digits generates the same width
        render_str += " " * 2 * (5 - len(str(self.chunk_count)))

        # Add extra spaces to align with the speed string
        render_str += " " * 6

        # Add a fixed-length, right-aligned percentage meter to align with the speed
        formatter_length = (
                6
                + 2 * len(self.progress_tracker.style_text_total) + len("[][/]")
                + 2 * len(self.progress_tracker.style_text_unit) + len("[][/]")
        )
        formatter = f"{{:>{formatter_length}s}}"
        render_str += formatter.format(
            f"[{self.progress_tracker.style_text_total}]"
            f"{100 * self.row_task.completed / self.row_task.total if self.row_task.total else 0:.1f}"
            f"[/{self.progress_tracker.style_text_total}]"
            f"[{self.progress_tracker.style_text_unit}]"
            f"%[/{self.progress_tracker.style_text_unit}]")

        return render_str


class _RowProgressBar(rich.progress.BarColumn):
    """Display a progress bar based on the current row completion and total row count.
    """

    def __init__(self, progress_tracker: ProgressTracker):
        super().__init__(bar_width=None)
        self.progress_tracker = progress_tracker
        self.complete_style = progress_tracker.style_bar_complete
        self.finished_style = progress_tracker.style_bar_finished
        self.style = progress_tracker.style_bar_incomplete
        self.pulse_style = "#0000ff"

    def render(self, task):
        return super().render(self.progress_tracker.row_task)


class _SeparatorColumn(_ProgressTextColumn):
    """Display an empty string.
    """

    def get_render_str(self) -> str:
        return ""



import enb.plugins


class ACPlugin(enb.plugins.PluginMake):
    name = "arithmetic_codec"
    label = "Arithmetic codec (8 bit)"
    tags = {"data compression", "codec"}
    contrib_authors = ["Nayuki"]
    contrib_reference_urls = ["https://github.com/nayuki/Reference-arithmetic-coding/"]
    contrib_download_url_name = [
        (
        "https://github.com/miguelinux314/experiment-notebook/blob/dev/contrib/nayuki_arithmetic_coder_reference.zip?raw=true",
        "nayuki_arithmetic_coder_reference.zip")]
    tested_on = {"linux"}



import enb.plugins


class SpeckPlugin(enb.plugins.plugin.PluginJava):
    name = "speck"
    label = "Wrapper for the SPECK codec"
    tags = {"data compression", "image", "codec"}
    contrib_authors = ["Group on Interactive Coding of Images (GICI)"]
    contrib_reference_urls = [
        "https://github.com/miguelinux314/experiment-notebook/blob/dev/contrib/SPModule_20220804.zip?raw=true"]
    tested_on = {"linux"}



#!/usr/bin/env python3
"""Tools to execute functions in remote using the ray library
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2019/11/21"

import os
import sys
import glob
import logging
import math
import threading
import time
import builtins
import ast
import shutil
import string
import subprocess
import random
import signal
import platform
import importlib
import psutil
import pandas as pd
import textwrap
import itertools
import contextlib

import enb
from . import config
from .config import options
from . import log
from .log import logger

# Ray is only expected to be available on linux nodes when clustering is desired.
# pylint: disable=invalid-name
try:
    import ray

    _ray_present = True
except ImportError as ex:
    _ray_present = False

# Flags indicating whether external tools are present
_ssh_present = shutil.which("ssh") is not None
_sshfs_present = shutil.which("sshfs") is not None
_dpipe_present = shutil.which("dpipe") is not None
_ray_cli_present = shutil.which("ray") is not None
_ray_disabled_warning_issued = False


def is_ray_enabled():
    """Return True if and only if ray is available and the current platform
    is one of the supported for ray clustering (currently only linux).
    """
    # pylint: disable=global-statement
    global _ray_disabled_warning_issued

    # ray is only enabled if an ssh cluster configuration file is provided
    if not options.ssh_cluster_csv_path:
        return False

    # ray is disabled if the ray command cannot be found
    if not _ray_present:
        if not _ray_disabled_warning_issued:
            enb.logger.warn(
                "An enb cluster configuration was selected "
                f"({repr(options.ssh_cluster_csv_path)}) "
                "but 'ray' could not be found in the path. "
                "Please install with `pip install ray[default]` "
                "and/or fix the path")
            _ray_disabled_warning_issued = True
        return False

    # ray is disabled on windows (it does not seem to work)
    if platform.system().lower() == "windows":
        if not _ray_disabled_warning_issued:
            enb.logger.warn(
                "An enb cluster configuration was selected "
                f"({repr(options.ssh_cluster_csv_path)}) "
                "but ray is not currently supported on Windows.")
            _ray_disabled_warning_issued = True
        return False

    # ray is disabled in any of the needed tools are missing
    failing_tool = None
    needed_package = None
    if not _ssh_present:
        failing_tool = "ssh"
    elif not _sshfs_present:
        failing_tool = "sshfs" if not options.no_remote_mount_needed else None
    elif not _dpipe_present:
        failing_tool = "dpipe" if not options.no_remote_mount_needed else None
        needed_package = "vde2"
    if failing_tool:
        if not _ray_disabled_warning_issued:
            enb.logger.warn(
                f"An enb cluster configuration was selected "
                f"({repr(options.ssh_cluster_csv_path)}) "
                f"but {failing_tool} was not found in the path. "
                f"No remote nodes will be used in this session.\n"
                f"Please install "
                f"{needed_package if needed_package is not None else failing_tool} "
                f"in your system and/or fix the path it and retry.")
            _ray_disabled_warning_issued = True
        return False

    return True


class HeadNode:
    """Class used to initialize and stop a ray head node.

    The stop() method must be called after start(), or a ray cluster will
    remain active.
    """

    def __init__(self, ray_port, ray_port_count):
        if not is_ray_enabled():
            raise RuntimeError(
                "The ray module is not present or is not available. "
                "Please see "
                "https://miguelinux314.github.io/experiment-notebook/cluster_setup.html "
                "for information about how to configure it.")

        assert ray_port == int(ray_port), ray_port
        assert ray_port_count == int(ray_port_count), ray_port_count
        assert 1025 <= ray_port, ray_port
        assert ray_port_count >= 1
        assert ray_port + ray_port_count - 1 <= 65535, (
            ray_port, ray_port_count, ray_port + ray_port_count - 1)
        self.ray_port = int(ray_port)
        self.ray_port_count = int(ray_port_count)
        self.session_password = ''.join(
            random.choices(string.ascii_letters, k=128))
        # List of RemoteNode instances started by this head node
        self.remote_nodes = []
        self.address = self.get_node_ip()

    def start(self):
        """Start or restart a ray head node.
        """
        with logger.info_context("Stoping any previous instance of ray..."):
            invocation = "ray stop --force"
            status, output = subprocess.getstatusoutput(invocation)
            if status != 0:
                raise Exception(f"Status = {status} != 0.\n"
                                f"Input={repr(invocation)}].\n"
                                f"Output={repr(output)}")

        with logger.info_context(f"Starting ray on port {self.ray_port}"):
            invocation = f"ray start --head " \
                         f"--include-dashboard false " \
                         f"--port {self.ray_port} " \
                         f"--ray-client-server-port {self.ray_port + 1} " \
                         f"--node-manager-port {self.ray_port + 2} " \
                         f"--object-manager-port {self.ray_port + 3} " \
                         f"--gcs-server-port  {self.ray_port + 4} " \
                         f"--min-worker-port  {self.ray_port + 5} " \
                         f"--max-worker-port  {self.ray_port + self.ray_port_count - 1} " \
                         f"--redis-password='{self.session_password}' " \
                         + (
                             f" --num-cpus {options.cpu_limit}" if options.cpu_limit else "")
            status, output = subprocess.getstatusoutput(invocation)
            if status != 0:
                raise RuntimeError(f"Error starting head ray process\n"
                                   f"Command: {repr(invocation)}.\n"
                                   f"Output:{repr(output)}")

        with logger.info_context(f"Initializing ray client on local port {self.ray_port}"):
            # pylint: disable=protected-access
            # A list of modules imported after initializing enb is passed to the
            # remote workers so that they can replicate the imports.
            modules_needed_remotely = [
                m.__name__ for m in sys.modules.values()
                if hasattr(m, "__name__")
                   and m.__name__ not in options._initial_module_names
                   and not m.__name__.startswith("_")]

            ray.init(address=f"localhost:{self.ray_port}",
                     _redis_password=self.session_password,
                     runtime_env=dict(
                         env_vars=dict(
                             # List of modules to import when enb is imported remotely
                             _needed_modules=str(modules_needed_remotely),
                             # Easy way to determine whether a worker is on a remote node or not
                             _head_node_ip=enb.misc.get_node_ip(),
                             # Add the remotely mounted project to sys.path so that ray
                             # can perform adequate deserialization
                             PYTHONPATH=str(RemoteNode.remote_project_mount_path
                                            if options.ssh_cluster_csv_path else options.project_root),
                         ),
                         # Workers need to chdir to the remotely mounted dir so that
                         # all imports work as expected
                         worker_process_setup_hook=lambda: os.chdir(RemoteNode.remote_project_mount_path),
                     ),
                     logging_level=logging.CRITICAL)

        if options.ssh_cluster_csv_path:
            failing_tool = None
            needed_package = None
            if not _ssh_present:
                failing_tool = "ssh"
            elif not _sshfs_present:
                failing_tool = "sshfs" if not options.no_remote_mount_needed else None
            elif not _dpipe_present:
                failing_tool = "dpipe" if not options.no_remote_mount_needed else None
                needed_package = "vde2"
            if failing_tool:
                requirements = needed_package if needed_package is not None \
                    else failing_tool
                enb.logger.warn(
                    f"An enb cluster configuration was selected "
                    f"({repr(options.ssh_cluster_csv_path)}) "
                    f"but {failing_tool} was not found in the path. "
                    f"No remote nodes will be used in this session.\n"
                    f"Please install {requirements} "
                    f"in your system and/or fix the path it and retry.")
            else:
                if not os.path.exists(options.ssh_cluster_csv_path):
                    raise ValueError(
                        "The cluster configuration file was set to "
                        f"{repr(options.ssh_cluster_csv_path)} "
                        "but it does not exist. "
                        "Either set enb.config.options.ssh_cluster_csv_path to None "
                        "or to an existing file. See "
                        "https://miguelinux314.github.io/experiment-notebook/installation.html"
                        " for more details.")

                self.remote_nodes = self.parse_cluster_config_csv(options.ssh_cluster_csv_path)
                logger.info(f"Connecting {len(self.remote_nodes)} remote nodes.")
                connected_nodes = []
                for rn in self.remote_nodes:
                    logger.info(f"Connecting to {rn}")
                    try:
                        rn.connect()
                        connected_nodes.append(rn)
                    except RuntimeError as connection_ex:
                        reason_str = textwrap.indent("\n".join(
                            itertools.chain(
                                *(textwrap.wrap(line, shutil.get_terminal_size()[0] - 6)
                                  for line in
                                  repr(connection_ex).replace(r"\n", "\n").splitlines()))),
                            " " * 4)

                        logger.warn(
                            f"Cannot connect to {rn.address}:{rn.ssh_port}. Reason:\n" +
                            reason_str +
                            f".\nExecution will continue without node {rn.address}:{rn.ssh_port}.")
                self.remote_nodes = connected_nodes
                logger.info(f"Done connecting {len(self.remote_nodes)} remote nodes.")

    def stop(self):
        """Stop the ray head node after disconnecting from all remote nodes.
        """
        with logger.info_context("Disconnecting from ray"):
            ray.shutdown()

        if self.remote_nodes:
            with logger.info_context("Stopping remote nodes...\n",
                                     msg_after="disconnected all remote nodes."):
                for rn in self.remote_nodes:
                    rn.disconnect()
            self.remote_nodes = []

        with logger.info_context("Stopping ray server."):
            # This tiny delay allows error messages from child processes to reach the
            # orchestrating process for logging.
            # It might need to be tuned for distributed computation across networks.
            time.sleep(options.preshutdown_wait_seconds)

            invocation = "ray stop --force"
            status, output = subprocess.getstatusoutput(invocation)
            if status != 0:
                logger.error(
                    "Error stopping ray process. You might need to run `ray stop` manually.\n"
                    f"Command: {repr(invocation)}.\nOutput: {repr(output)}")

    def parse_cluster_config_csv(self, csv_path):
        """Read a CSV defining remote nodes and return a list with as many
        RemoteNode as data rows in the CSV.
        """

        def clean_value(s, default_value=None):
            return s if (isinstance(s, str) and s) or not math.isnan(s) else default_value

        return [RemoteNode(
            address=clean_value(row["address"], None),
            ssh_user=clean_value(row["ssh_user"], os.getlogin()),
            ssh_port=clean_value(row["ssh_port"], 22),
            local_ssh_file=clean_value(row["local_ssh_file"], None),
            cpu_limit=clean_value(row["cpu_limit"], None),
            head_node=self)
            for _, row in pd.read_csv(csv_path, comment="#").iterrows()]

    def get_node_ip(self):
        """Adapted from https://stackoverflow.com/a/166589/992926.
        """
        assert not is_parallel_process()

        try:
            return self._head_node_address
        except AttributeError:
            self._head_node_address = enb.misc.get_node_ip()
            return self._head_node_address

    @property
    def status_str(self):
        """Return a string reporting the status of the cluster"""
        cpu_str = f"{int(ray.cluster_resources()['CPU'])} virtual CPU cores"
        gpu_str = str(int(ray.cluster_resources()['GPU'])
                      if 'GPU' in ray.cluster_resources() else 0) + \
                  " GPU devices"
        remote_node_str = "\n\t\t * ".join(str(rn) for rn in self.remote_nodes)

        return f"The current enb/ray cluster consists of:\n" \
               f"\t- {len(self.remote_nodes) + 1} total nodes.\n" + \
            ((f"\t- {len(self.remote_nodes)} remote nodes:"
              f"\n\t\t * {remote_node_str}\n") if self.remote_nodes else "") + \
            f"\t- {cpu_str}.\n" \
            f"\t- {gpu_str}."


# Single HeadNode instance that controls the ray cluster
_head_node = None


class RemoteNode:
    """Represent a remote node of the cluster, with tools to connect via ssh.
    """
    # pylint: disable=too-many-instance-attributes
    remote_project_mount_path = os.path.join(enb.user_config_dir, "remote_mount")

    def __init__(self, address, ssh_port, head_node, ssh_user=None,
                 local_ssh_file=None, cpu_limit=None,
                 remote_mount_needed=None):
        # pylint: disable=too-many-arguments
        assert is_ray_enabled()

        self.address = address
        self.ssh_user = ssh_user
        self.ssh_port = ssh_port
        self.local_ssh_file = local_ssh_file
        self.head_node = head_node
        self.mount_popen = None
        self.cpu_limit = cpu_limit
        if self.cpu_limit is not None and self.cpu_limit <= 0:
            self.cpu_limit = None
        self.remote_mount_needed = not options.no_remote_mount_needed \
            if remote_mount_needed is None else remote_mount_needed

    def connect(self):
        """Connect to a remote ray head node.
        """
        assert not is_parallel_process()

        ssh_address = \
            f"{self.ssh_user + '@' if self.ssh_user else ''}{self.address}"

        # Create remote_node_folder_path on the remote host if not existing
        with logger.info_context(f"Stopping ray on {self.address}"):
            invocation = \
                f"ssh -p {self.ssh_port if self.ssh_port else 22} " \
                f"-i {self.local_ssh_file if self.local_ssh_file else ''} " \
                f"{ssh_address} " \
                f"ray stop --force"
            status, output = subprocess.getstatusoutput(invocation)
            if status != 0:
                raise RuntimeError(
                    f"Error stopping remote ray process on {self}.\n"
                    f"Command: {repr(invocation)}.\nOutput: {repr(output)}")

        if self.remote_mount_needed:
            # Create remote_node_folder_path on the remote host if not existing
            with logger.info_context(
                    f"Creating remote mount point on {self.address}"):
                invocation = f"ssh -p {self.ssh_port if self.ssh_port else 22} " \
                             f"{'-i ' + self.local_ssh_file if self.local_ssh_file else ''} " \
                             f"{self.ssh_user + '@' if self.ssh_user else ''}{self.address} " \
                             f"mkdir -p {self.remote_project_mount_path}"
                status, output = subprocess.getstatusoutput(invocation)
                if status != 0:
                    raise RuntimeError(
                        f"Error creating remote mount point on {self}.\n"
                        f"Command: {repr(invocation)}.\nOutput: {repr(output)}")

            # Umount remote_node_folder_path on the remote host if previously mounted
            with logger.info_context(
                    f"Unmounting remote mount point on {self.address} (if it was mounted)"):
                invocation = f"ssh -p {self.ssh_port if self.ssh_port else 22} " \
                             f"{'-i ' + self.local_ssh_file if self.local_ssh_file else ''} " \
                             f"{self.ssh_user + '@' if self.ssh_user else ''}{self.address} " \
                             f"umount {self.remote_project_mount_path}"
                subprocess.getstatusoutput(invocation)

            # Mount the project root on remote_node_folder_path - use a separate process
            self.mount_project_remotely()
        else:
            enb.logger.debug(f"{self}: Not mounting via sshfs because "
                             f"self.mount_remotely={self.remote_mount_needed}")

        with logger.info_context(f"Starting ray process on {self.address}"):
            invocation = \
                f"ssh -p {self.ssh_port if self.ssh_port else 22} " \
                f"{'-i ' + self.local_ssh_file if self.local_ssh_file else ''} " \
                f"{self.ssh_user + '@' if self.ssh_user else ''}{self.address} " \
                f"ray start --address " \
                f"{self.head_node.address}:{self.head_node.ray_port} " \
                f"--ray-client-server-port {self.head_node.ray_port + 1} " \
                f"--node-manager-port {self.head_node.ray_port + 2} " \
                f"--object-manager-port {self.head_node.ray_port + 3} " \
                f"--min-worker-port {self.head_node.ray_port + 5} " \
                f"--max-worker-port " \
                f"{self.head_node.ray_port + self.head_node.ray_port_count - 1} " \
                f"--redis-password='{self.head_node.session_password}' " \
                + (f" --num-cpus {self.cpu_limit}" if self.cpu_limit else "")
            status, output = subprocess.getstatusoutput(invocation)
            if status != 0:
                raise RuntimeError(f"Error starting remote ray on {self}.\n"
                                   f"Command: {repr(invocation)}.\nOutput: {repr(output)}")

    def mount_project_remotely(self):
        """Use sshfs to mount the remote project folder into the remote node.
        """
        # Mount the project root on remote_node_folder_path
        invocation = f"dpipe /usr/lib/openssh/sftp-server = " \
                     f"ssh -p {self.ssh_port if self.ssh_port else 22} " \
                     f"{'-i ' + self.local_ssh_file if self.local_ssh_file else ''} " \
                     f"{self.ssh_user + '@' if self.ssh_user else ''}{self.address} " \
                     f"sshfs :{options.project_root} {self.remote_project_mount_path} " \
                     "-C -o sshfs_sync -o slave"

        # pylint: disable=consider-using-with,subprocess-popen-preexec-fn
        self.mount_popen = subprocess.Popen(
            invocation, stdout=subprocess.PIPE,
            preexec_fn=os.setsid, shell=True)
        # pylint: enable=consider-using-with,subprocess-popen-preexec-fn

        remote_mount_thread = threading.Thread(
            target=self.mount_popen.communicate, daemon=True)
        remote_mount_thread.start()

        remote_mount_thread.join(timeout=1)
        if not remote_mount_thread.is_alive():
            raise RuntimeError(
                "Error mounting project folder remotely "
                "(is sshfs installed in the remote node?)")

    def disconnect(self):
        """Disconnect from a remote node.
        """
        assert not is_parallel_process()

        # Create remote_node_folder_path on the remote host if not existing
        with logger.info_context(f"Disconnecting {self.address} (stopping ray)"):
            invocation = f"ssh -p {self.ssh_port if self.ssh_port else 22} " \
                         f"{'-i ' + self.local_ssh_file if self.local_ssh_file else ''} " \
                         f"{self.ssh_user + '@' if self.ssh_user else ''}{self.address} " \
                         f"ray stop --force"
            status, output = subprocess.getstatusoutput(invocation)
            if status != 0:
                logger.debug(f"Error disconnecting {self}. "
                             f"Command: {repr(invocation)}.\nOutput: {repr(output)}")

        if self.remote_mount_needed:
            target_str = f"{self.ssh_user + '@' if self.ssh_user else ''}{self.address} " \
                         f"sshfs :{options.project_root}"
            for proc in psutil.process_iter():
                cmd_str = " ".join(proc.cmdline())
                if target_str in cmd_str:
                    try:
                        os.kill(proc.pid, signal.SIGTERM)
                    except ProcessLookupError:
                        logger.info(
                            f"Cannot kill previously found process {proc.pid}")

    def __repr__(self):
        return f"{self.__class__.__name__}(address={self.address}, " \
               f"ssh_port={self.ssh_port}, ssh_user={self.ssh_user}, " \
               f"local_ssh_file={self.local_ssh_file}, " \
               f"cpu_limit={self.cpu_limit})"


def init_ray():
    """Initialize the ray cluster if it wasn't initialized before.
    """
    # pylint: disable=global-statement
    global _head_node

    if not ray.is_initialized():
        if _head_node is not None:
            _head_node.stop()

        # Initialize cluster of workers
        logger.info(f"Initializing ray cluster [CPUlimit={options.cpu_limit}]...")
        if not options.disable_swap:
            # From https://github.com/ray-project/ray/issues/10895
            # - allow using swap memory when needed,
            # avoiding early termination of jobs due to that.
            os.environ["RAY_DEBUG_DISABLE_MEMORY_MONITOR"] = "1"

        _head_node = HeadNode(ray_port=options.ray_port,
                              ray_port_count=options.ray_port_count)
        _head_node.start()
        options.head_address = _head_node.get_node_ip()
        logger.verbose(_head_node.status_str + "\n")


def stop_ray():
    """Stop the ray head node, if one is defined.
    """
    if _head_node is not None:
        _head_node.stop()


def is_parallel_process():
    """Return True if and only if the call is made from a remote ray process,
    which can be running in the head node or any of the remote nodes (if any
    is present).
    """
    return os.path.basename(sys.argv[0]) == os.path.basename(
        options.worker_script_name)


def is_remote_node():
    """Return True if and only if the call is performed from a remote ray
    process running on a node different from the head.
    """
    if not is_parallel_process():
        return False
    try:
        if os.environ['_head_node_ip'] == enb.misc.get_node_ip():
            return False
    except KeyError:
        return False
    
    return True


def is_ray_initialized():
    """Return True if and only if ray is enabled and initialized.
    """
    return is_ray_enabled() and ray.is_initialized


def parallel_decorator(*args, **kwargs):
    """Wrapper of the @`ray.remote` decorator that automatically updates
    enb.config.options for remote processes, so that they always access the
    intended configuration.
    """
    kwargs["num_cpus"] = kwargs["num_cpus"] if "num_cpus" in kwargs else 1
    kwargs["num_gpus"] = kwargs["num_gpus"] if "num_gpus" in kwargs else 0

    def ray_remote_wrapper(f):
        enb.logger.debug(f"Wrapping {f} with ray")

        def remote_method_wrapper(_opts, *a, **k):
            """Wrapper for the decorated function f, that updates
            enb.config.options before f is called.
            """
            config.options.update(_opts, trigger_events=False)
            new_level = logger.get_level(logger.level_message.name,
                                         config.options.verbose)
            log.logger.selected_log_level = new_level
            builtins.print = logger.print_to_log

            return f(*a, **k)

        method_proxy = ray.remote(*args, **kwargs)(remote_method_wrapper)
        method_proxy.ray_remote = method_proxy.remote

        def local_side_remote(*a, **k):
            """Wrapper for ray's `.parallel_decorator()` method invoked in
            the local side. It makes sure that `remote_side_wrapper` receives
            the options argument.
            """
            # apply ray.put to all arguments before passing them
            args = []
            args.append(ray.put(dict(config.options.items())))
            args.extend(ray.put(argument) for argument in a)
            kwargs = {key: ray.put(value) for key, value in k.items()}
            with (open(os.devnull, "w") as devnull,
                  contextlib.redirect_stdout(devnull)):
                return method_proxy.ray_remote(*args, **kwargs)

        del method_proxy.remote
        method_proxy.start = local_side_remote

        return method_proxy

    return ray_remote_wrapper


def get(ids, **kwargs):
    """Call ray's get method with the given arguments.
    """
    return ray.get(ids, **kwargs)


def get_completed_pending_ids(ids, timeout=0):
    """Return the list of completed and pending ids.
    """
    return ray.wait(ids, num_returns=len(ids), timeout=timeout)

def chdir_project_root():
    """When invoked, it changes the current working dir to the project's root. It will be
    the remote mount point if the node is remote, otherwise the directory containing
    the invoking script.
    """
    if is_remote_node():
        os.chdir(os.path.expanduser(RemoteNode.remote_project_mount_path))
    elif not enb.is_enb_cli:
        os.chdir(options.project_root)

def fix_imports():
    """An environment variable is passed to the children processes for them
    to be able to import all modules that were imported after loading enb.
    This prevents the remote functions to fail the deserialization process
    due to missing definitions.
    """
    chdir_project_root()

    if is_remote_node():
        imported_modules = set()
        for module_name in sorted(
                ast.literal_eval(os.environ['_needed_modules'])):
            try:
                importlib.import_module(module_name)
                imported_modules.add(module_name)
            except ImportError as _ex:
                module_parts = module_name.split(".")
                for i in range(1, len(module_parts) - 1):
                    if ".".join(module_parts[:i]) in imported_modules:
                        break
                    logger.error(f"Error importing module {repr(module_name)}: {repr(_ex)}.")



#!/usr/bin/env python3
"""Image compression experiment module.
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2020/04/01"

import os
import re
import filecmp
import hashlib
import tempfile
import time
import collections
import functools
import shutil
import math
import numpy as np
from scipy import signal
from scipy.ndimage.filters import convolve

from enb.atable import ColumnProperties, column_function, indices_to_internal_loc
from enb.isets import load_array_bsq, dump_array_bsq

# These * imports provide backwards import compatibility for versions < v1.1.0,
# but enb.compression.* is preferred since v1.1.0.
from enb.compression import * 
from enb.compression.codec import *
from enb.compression.wrapper import *


class LossyCompressionExperiment(CompressionExperiment):
    """Lossy compression of raw image files.
    """

    @column_function("mse", label="MSE", plot_min=0)
    def set_MSE(self, index, row):
        """Set the mean squared error of the reconstructed image.
        """
        original_array = np.fromfile(
            self.codec_results.compression_results.original_path,
            dtype=self.codec_results.numpy_dtype).astype(np.int64)

        reconstructed_array = np.fromfile(
            self.codec_results.decompression_results.reconstructed_path,
            dtype=self.codec_results.numpy_dtype).astype(np.int64)
        row[_column_name] = np.average(
            ((original_array - reconstructed_array) ** 2))

    @column_function("pae", label="PAE", plot_min=0)
    def set_PAE(self, index, row):
        """Set the peak absolute error (maximum absolute pixelwise
        difference) of the reconstructed image.
        """
        original_array = np.fromfile(
            self.codec_results.compression_results.original_path,
            dtype=self.codec_results.numpy_dtype).astype(np.int64)

        reconstructed_array = np.fromfile(
            self.codec_results.decompression_results.reconstructed_path,
            dtype=self.codec_results.numpy_dtype).astype(np.int64)
        row[_column_name] = np.max(np.abs(original_array - reconstructed_array))

    @column_function("psnr_bps", label="PSNR (dB)", plot_min=0)
    def set_PSNR_nominal(self, index, row):
        """Set the PSNR assuming nominal dynamic range given by
        bytes_per_sample.
        """
        file_path, codec_name = self.index_to_path_task(index)
        row.image_info_row = self.dataset_table_df.loc[indices_to_internal_loc(file_path)]
        if row.image_info_row["float"]:
            row[_column_name] = float("inf")
        else:
            max_error = (2 ** (8 * row.image_info_row["bytes_per_sample"])) - 1
            row[_column_name] = (
                    20 * math.log10((max_error) / math.sqrt(row["mse"]))) \
                if row["mse"] > 0 else float("inf")

    @column_function("psnr_dr", label="PSNR (dB)", plot_min=0)
    def set_PSNR_dynamic_range(self, index, row):
        """Set the PSNR assuming dynamic range given by dynamic_range_bits.
        """
        file_path, codec_name = self.index_to_path_task(index)
        row.image_info_row = self.dataset_table_df.loc[
            indices_to_internal_loc(file_path)]
        max_error = (2 ** row.image_info_row["dynamic_range_bits"]) - 1
        row[_column_name] = 20 * math.log10(max_error / math.sqrt(row["mse"])) \
            if row["mse"] > 0 else float("inf")


class StructuralSimilarity(CompressionExperiment):
    """Set the Structural Similarity (SSIM) and Multi-Scale Structural
    Similarity metrics (MS-SSIM) to measure the similarity between two images.

    Authors:
        - http://www.cns.nyu.edu/~lcv/ssim/msssim.zip
        - https://github.com/dashayushman/TAC-GAN/blob/master/msssim.py
    """

    @column_function([
        ColumnProperties(name="ssim", label="SSIM", plot_max=1),
        ColumnProperties(name="ms_ssim", label="MS-SSIM",
                         plot_max=1)])
    def set_StructuralSimilarity(self, index, row):
        file_path, codec_name = self.index_to_path_task(index)
        row.image_info_row = self.dataset_table_df.loc[
            indices_to_internal_loc(file_path)]
        original_array = np.fromfile(
            self.codec_results.compression_results.original_path,
            dtype=self.codec_results.numpy_dtype)
        original_array = np.reshape(original_array,
                                    (row.image_info_row["width"],
                                     row.image_info_row["height"],
                                     row.image_info_row["component_count"]))

        reconstructed_array = np.fromfile(
            self.codec_results.decompression_results.reconstructed_path,
            dtype=self.codec_results.numpy_dtype)
        reconstructed_array = np.reshape(reconstructed_array,
                                         (row.image_info_row["width"],
                                          row.image_info_row["height"],
                                          row.image_info_row[
                                              "component_count"]))

        row["ssim"] = self.compute_SSIM(original_array, reconstructed_array)
        row["ms_ssim"] = self.cumpute_MSSIM(original_array, reconstructed_array)

    def cumpute_MSSIM(self, img1, img2, max_val=255, filter_size=11,
                      filter_sigma=1.5, k1=0.01, k2=0.03, weights=None):
        """Return the MS-SSIM score between `img1` and `img2`.

        This function implements Multi-Scale Structural Similarity (MS-SSIM) Image
        Quality Assessment according to Zhou Wang's paper, "Multi-scale structural
        similarity for image quality assessment" (2003).
        Link: https://ece.uwaterloo.ca/~z70wang/publications/msssim.pdf

        Author's MATLAB implementation:
        http://www.cns.nyu.edu/~lcv/ssim/msssim.zip

        Author's Python implementation:
        https://github.com/dashayushman/TAC-GAN/blob/master/msssim.py

        Authors documentation:

        :param img1: Numpy array holding the first RGB image batch.
        :param img2: Numpy array holding the second RGB image batch.
        :param max_val: the dynamic range of the images (i.e., the difference
          between the maximum the and minimum allowed values).
        :param filter_size: Size of blur kernel to use (will be reduced for
          small images).
        :param filter_sigma: Standard deviation for Gaussian blur kernel (
          will be reduced for small images).
        :param k1: Constant used to maintain stability in the SSIM
          calculation (0.01 in the original paper).
        :param k2: Constant used to maintain stability in the SSIM
          calculation (0.03 in the original paper).
        """
        # pylint: disable=too-many-arguments
        if img1.shape != img2.shape:
            raise RuntimeError(
                'Input images must have the same shape (%s vs. %s).',
                img1.shape, img2.shape)
        if img1.ndim != 3:
            raise RuntimeError("Input images must have four dimensions, "
                               f"not {img1.ndim}")

        weights = np.array(weights if weights else
                           [0.0448, 0.2856, 0.3001, 0.2363, 0.1333])
        levels = weights.size
        downsample_filter = np.ones((2, 2, 1)) / 4.0
        im1, im2 = [x.astype(np.float64) for x in [img1, img2]]
        mssim = np.array([])
        mcs = np.array([])
        for _ in range(levels):
            ssim, cs = self.compute_SSIM(
                im1, im2, max_val=max_val, filter_size=filter_size,
                filter_sigma=filter_sigma, k1=k1, k2=k2, full=True)
            mssim = np.append(mssim, ssim)
            mcs = np.append(mcs, cs)
            filtered = [convolve(im, downsample_filter, mode='reflect')
                        for im in [im1, im2]]
            im1, im2 = [x[::2, ::2, :] for x in filtered]

        return np.prod(mcs[0:levels - 1] ** weights[0:levels - 1]) * (
                mssim[levels - 1] ** weights[levels - 1])

    def compute_SSIM(self, img1, img2, max_val=255, filter_size=11,
                     filter_sigma=1.5, k1=0.01, k2=0.03, full=False):
        """Return the Structural Similarity Map between `img1` and `img2`.

        This function attempts to match the functionality of ssim_index_new.m
        by Zhou Wang: http://www.cns.nyu.edu/~lcv/ssim/msssim.zip

        Author's Python implementation:
        https://github.com/dashayushman/TAC-GAN/blob/master/msssim.py

        :param img1: Numpy array holding the first RGB image batch.
        :param img2: Numpy array holding the second RGB image batch.

        :param max_val: the dynamic range of the images (i.e., the difference
          between the maximum the and minimum allowed values).

        :param filter_size: Size of blur kernel to use (will be reduced for
          small images). :param filter_sigma: Standard deviation for Gaussian
          blur kernel (will be reduced for small images). :param k1: Constant
          used to maintain stability in the SSIM calculation (0.01 in the
          original paper). :param k2: Constant used to maintain stability in
          the SSIM calculation (0.03 in the original paper).
        """
        # pylint: disable=too-many-arguments
        if img1.shape != img2.shape:
            raise RuntimeError(
                'Input images must have the same shape (%s vs. %s).',
                img1.shape, img2.shape)
        if img1.ndim != 3:
            raise RuntimeError(
                f"Input images must have four dimensions, not {img1.ndim}")
        img1 = img1.astype(np.float64)
        img2 = img2.astype(np.float64)
        height, width, bytes = img1.shape

        # Filter size can't be larger than height or width of images.
        size = min(filter_size, height, width)

        # Scale down sigma if a smaller filter size is used.
        sigma = size * filter_sigma / filter_size if filter_size else 0

        if filter_size:
            window = np.reshape(self._FSpecialGauss(size, sigma),
                                (size, size, 1))
            mu1 = signal.fftconvolve(img1, window, mode='valid')
            mu2 = signal.fftconvolve(img2, window, mode='valid')
            sigma11 = signal.fftconvolve(img1 * img1, window, mode='valid')
            sigma22 = signal.fftconvolve(img2 * img2, window, mode='valid')
            sigma12 = signal.fftconvolve(img1 * img2, window, mode='valid')
        else:
            # Empty blur kernel so no need to convolve.
            mu1, mu2 = img1, img2
            sigma11 = img1 * img1
            sigma22 = img2 * img2
            sigma12 = img1 * img2

        mu11 = mu1 * mu1
        mu22 = mu2 * mu2
        mu12 = mu1 * mu2
        sigma11 -= mu11
        sigma22 -= mu22
        sigma12 -= mu12

        # Calculate intermediate values used by both ssim and cs_map.
        c1 = (k1 * max_val) ** 2
        c2 = (k2 * max_val) ** 2
        v1 = 2.0 * sigma12 + c2
        v2 = sigma11 + sigma22 + c2
        ssim = np.mean((((2.0 * mu12 + c1) * v1) / ((mu11 + mu22 + c1) * v2)))
        cs = np.mean(v1 / v2)
        if full:
            return ssim, cs
        return ssim

    def _FSpecialGauss(self, size, sigma):

        radius = size // 2
        offset = 0.0
        start, stop = -radius, radius + 1
        if size % 2 == 0:
            offset = 0.5
            stop -= 1
        x, y = np.mgrid[offset + start:stop, offset + start:stop]
        assert len(x) == size
        g = np.exp(-((x ** 2 + y ** 2) / (2.0 * sigma ** 2)))
        return g / g.sum()


class SpectralAngleTable(LossyCompressionExperiment):
    """Lossy compression experiment that computes spectral angle "distance"
    measures between the compressed and the reconstructed images.

    Subclasses of LossyCompressionExperiment may inherit from this one to
    automatically add the data columns defined here
    """

    def get_spectral_angles_deg(self, index, row):
        """Return a sequence of spectral angles (in degrees),
        one per (x,y) position in the image, flattened in raster order.
        """
        # Read original and reconstructed images
        original_file_path, task_name = index
        image_properties_row = self.get_dataset_info_row(original_file_path)
        decompression_results = self.codec_results.decompression_results
        original_array = load_array_bsq(
            file_or_path=original_file_path,
            image_properties_row=image_properties_row)
        reconstructed_array = load_array_bsq(
            file_or_path=decompression_results.reconstructed_path,
            image_properties_row=image_properties_row)

        # Reshape flattening the x,y axes, and maintaining the z axis for
        # each (x,y) position
        original_array = np.reshape(
            original_array.swapaxes(0, 1),
            (image_properties_row["width"] * image_properties_row["height"],
             image_properties_row["component_count"]),
            "F").astype("i8")
        reconstructed_array = np.reshape(
            reconstructed_array.swapaxes(0, 1),
            (image_properties_row["width"] * image_properties_row["height"],
             image_properties_row["component_count"]),
            "F").astype("i8")

        dots = np.einsum("ij,ij->i", original_array, reconstructed_array)
        magnitude_a = np.linalg.norm(original_array, axis=1)
        magnitude_b = np.linalg.norm(reconstructed_array, axis=1)

        for i in range(magnitude_a.shape[0]):
            # Avoid division by zero
            magnitude_a[i] = max(1e-4, magnitude_a[i])
            magnitude_b[i] = max(1e-4, magnitude_b[i])

        # Clip, because the dot product can slip past 1 or -1 due to rounding
        cosines = np.clip(dots / (magnitude_a * magnitude_b), -1, 1)
        angles = np.degrees(np.arccos(cosines))
        # Round because two identical images should return an angle of exactly 0
        angles = np.round(angles, 5)

        return angles.tolist()

    @column_function([
        ColumnProperties("mean_spectral_angle_deg",
                         label="Mean spectral angle (deg)",
                         plot_min=0, plot_max=None),
        ColumnProperties("max_spectral_angle_deg",
                         label="Max spectral angle (deg)",
                         plot_min=0, plot_max=None)])
    def set_spectral_distances(self, index, row):
        spectral_angles = self.get_spectral_angles_deg(index=index, row=row)

        for angle in spectral_angles:
            assert not np.isnan(
                angle), f"Error calculating an angle for {index}: {angle}"

        row["mean_spectral_angle_deg"] = sum(spectral_angles) / len(
            spectral_angles)
        row["max_spectral_angle_deg"] = max(spectral_angles)


class GiciLibHelper:
    """Definition of helper methods that can be used with software based on
    the GiciLibs (see gici.uab.cat/GiciWebPage/downloads.php).
    """

    def file_info_to_data_str(self, original_file_info):
        if original_file_info["bytes_per_sample"] == 1:
            data_type_str = "1"
        elif original_file_info["bytes_per_sample"] == 2:
            if original_file_info["signed"]:
                data_type_str = "3"
            else:
                data_type_str = "2"
        elif original_file_info["bytes_per_sample"] == 4:
            if original_file_info["signed"]:
                return "4"
            else:
                raise ValueError(
                    "32-bit samples are supported, but they must be signed.")
        else:
            raise ValueError(
                f"Invalid data type, not supported by "
                f"{self.__class__.__name__}: {original_file_info}")
        return data_type_str

    def file_info_to_endianness_str(self, original_file_info):
        return "0" if original_file_info["big_endian"] else "1"

    def get_gici_geometry_str(self, original_file_info):
        """Get a string to be passed to the -ig or -og parameters. The '-ig'
        or '-og' part is not included in the returned string.
        """
        return f"{original_file_info['component_count']} " \
               f"{original_file_info['height']} " \
               f"{original_file_info['width']} " \
               f"{self.file_info_to_data_str(original_file_info=original_file_info)} " \
               f"{self.file_info_to_endianness_str(original_file_info=original_file_info)} " \
               f"0 "



#!/usr/bin/env python3
"""FSE codec wrapper
"""
__author__ = "Òscar Maireles and Miguel Hernández-Cabronero"
__since__ = "2021/06/01"

import os
from enb import icompression

class FSE(icompression.WrapperCodec, icompression.LosslessCodec):
    def __init__(self):
        icompression.WrapperCodec.__init__(
            self,
            compressor_path=os.path.join(os.path.dirname(os.path.abspath(__file__)), "fse"),
            decompressor_path=os.path.join(os.path.dirname(os.path.abspath(__file__)), "fse"),
            param_dict=None, output_invocation_dir=None)

    def get_compression_params(self, original_path, compressed_path, original_file_info):
        return f"-e -f {original_path} {compressed_path}"

    def get_decompression_params(self, compressed_path, reconstructed_path, original_file_info):
        return f"-e -d -f {compressed_path} {reconstructed_path}"

    @property
    def label(self):
        return "FSE"



import enb


class PlotGalleryTemplate(enb.plugins.Template):
    """Template for lossless and lossy data compression experiments.
    """
    name = "analysis-gallery"
    author = ["Miguel Hernández-Cabronero"]
    label = "Self-contained gallery of data analysis and plotting examples."
    tags = {"documentation"}
    tested_on = {"linux", "windows"}


#!/usr/bin/env python3
"""Codec wrapper for the Ndzip lossless image coder
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2020/05/25"

import os
import enb


class Ndzip(enb.icompression.LosslessCodec, enb.icompression.NearLosslessCodec, enb.icompression.WrapperCodec):
    """Wrapper for the Ndzip codec.
    Only float 16, 32 and 64 can be compressed .
    """

    def __init__(self, ndzip_binary=os.path.join(os.path.dirname(__file__), "compress")):

        super().__init__(compressor_path=ndzip_binary,
                         decompressor_path=ndzip_binary,
                         param_dict=dict())

    @property
    def label(self):
        return "ndzip"

    def get_compression_params(self, original_path, compressed_path, original_file_info):
        assert original_file_info["float"] == True, f"Only floating point data is supported by {self.label}"
        if original_file_info["bytes_per_sample"] == 8:
            return f"-t double -n {original_file_info['samples']} -i {original_path} -o {compressed_path}"
        elif original_file_info["bytes_per_sample"] == 4:
            return f"-t float -n {original_file_info['samples']} -i {original_path} -o {compressed_path}"
        else:
            raise ValueError(f"Only 32-bit and 64-bit float samples are supported by {self.label}")

    def get_decompression_params(self, compressed_path, reconstructed_path, original_file_info):
        assert original_file_info["float"] == True, f"Only floating point data is supported by {self.label}"
        if original_file_info["bytes_per_sample"] == 8:
            return f"-d -t double -n {original_file_info['samples']} -i {compressed_path} -o {reconstructed_path}"
        elif original_file_info["bytes_per_sample"] == 4:
            return f"-d -t float -n {original_file_info['samples']} -i {compressed_path} -o {reconstructed_path}"
        else:
            raise ValueError(f"Only 32-bit and 64-bit float samples are supported by {self.label}")



#!/usr/bin/env python3
"""Wrapper for the Terse/Prolix lossless codec.

https://github.com/senikm/trpx
https://senikm.github.io/trpx/
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2024/11/11"

import os

import enb.isets
from enb import icompression


class TRPX(icompression.WrapperCodec, icompression.LosslessCodec):
    """Wrapper for the Terse/Prolix lossless codec.
    https://github.com/senikm/trpx
    https://senikm.github.io/trpx/
    """

    def __init__(self, compressor_path=None, decompressor_path=None):
        if compressor_path is None:
            compressor_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "raw_codec")
        if decompressor_path is None:
            decompressor_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "raw_codec")
        super().__init__(compressor_path=compressor_path, decompressor_path=decompressor_path)

    def get_compression_params(self, original_path, compressed_path, original_file_info):
        geometry_dict = enb.isets.file_path_to_geometry_dict(
            file_path=original_path, existing_dict=dict(original_file_info))
        if geometry_dict["float"] is not False or geometry_dict["bytes_per_sample"] not in (1, 2, 4):
            raise ValueError(f"Invalid input data type; only uint8_t, uint16_t and uint32_t are supported,"
                             f"but the following was found for {original_path!r}: {geometry_dict}.")
        return f"compress {geometry_dict['bytes_per_sample']} {original_path} {compressed_path}"

    def get_decompression_params(self, compressed_path, reconstructed_path, original_file_info):
        geometry_dict = dict(original_file_info)
        if geometry_dict["float"] or geometry_dict["bytes_per_sample"] not in (1, 2, 4):
            raise ValueError(f"Invalid input data type; only uint8_t, uint16_t and uint32_t are supported,"
                             f"but the following was found: {geometry_dict}.\n\n"
                             f'@@{geometry_dict["float"]!r}@@')
        return f"decompress {geometry_dict['bytes_per_sample']} {compressed_path} {reconstructed_path}"

    @property
    def label(self):
        return "TRPX"



import textwrap
import enb.plugins


class NDZIPPlugin(enb.plugins.PluginMake):
    name = "ndzip"
    label = "Wrapper for ndzip"
    tags = {"data compression", "codec"}
    contrib_authors = ["Fabian Knorr"]
    contrib_reference_urls = ["https://github.com/fknorr/ndzip"]
    contrib_download_url_name = [
        ("https://github.com/miguelinux314/experiment-notebook/blob/dev/contrib/ndzip-master-enb.zip?raw=true",
         "ndzip-master-enb.zip")]
    extra_requirements_message = textwrap.dedent("""
    The following system packages should be installed for this plugin to be able to compile:
    
        - cmake
        - libboost-dev
        - libboost-thread-dev
        - libboost-program-options-dev
    """)
    tested_on = {"linux"}



import os
import zipfile
import enb.plugins


class EmpordaPlugin(enb.plugins.plugin.PluginJava):
    name = "emporda"
    label = "Wrapper for the emporda codec (AC and CCSDS versions)"
    tags = {"data compression", "image", "codec"}
    contrib_authors = ["Group on Interactive Coding of Images (GICI)"]
    contrib_reference_urls = ["http://gici.uab.cat/GiciWebPage/downloads.php#emporda"]
    contrib_download_url_name = [
        ("https://github.com/miguelinux314/experiment-notebook/blob/dev/contrib/"
         "emporda_20240419.zip?raw=true", "emporda.zip")]
    tested_on = {"linux"}

    @classmethod
    def build(cls, installation_dir):
        """Unzip the folder contents and build the plugin.
        """
        with zipfile.ZipFile(os.path.join(installation_dir, "emporda.zip"), "r") as zip_file:
            zip_file.extractall(installation_dir)
        super().build(installation_dir)



#!/usr/bin/env python3
"""Codec wrapper for the Fpack lossless and lossy image coder
"""
__author__ = "Òscar Maireles"

import os
import enb


class FPACK_Abstract(enb.icompression.LosslessCodec, enb.icompression.FitsWrapperCodec):
    def assert_valid_data_type(self, original_file_info):
        assert not original_file_info["float"], \
            f"Only integer samples are currently supported by {self.__class__.__name__}"
        assert original_file_info["big_endian"], \
            f"Only big-endian samples are currently supported by {self.__class__.__name__}"


class FPACK_GZIP(FPACK_Abstract):
    """fpack wrapper with gzip compression.
    """
    def __init__(self,
                 bin_dir=None):
        bin_dir = bin_dir if bin_dir is not None else os.path.dirname(__file__)
        super().__init__(compressor_path=os.path.join(bin_dir, "fpack"),
                         decompressor_path=os.path.join(bin_dir, "funpack"),
                         param_dict=dict())

    @property
    def name(self):
        """Don't include the binary signature
        """
        return f"{self.__class__.__name__}{'__' if self.param_dict else ''}" \
               f"{'_'.join(f'{k}={v}' for k, v in self.param_dict.items())}"

    @property
    def label(self):
        return "FPACK - GZIP"

    def get_compression_params(self, original_path, compressed_path, original_file_info):
        self.assert_valid_data_type(original_file_info=original_file_info)
        return f" -q 0 -s 0 -g   -O {compressed_path} {original_path}"

    def get_decompression_params(self, compressed_path, reconstructed_path, original_file_info):
        return f" -O {reconstructed_path} {compressed_path} "


class FPACK_RICE(FPACK_Abstract):
    """fpack wrapper with Rice compression.
    """
    def __init__(self,
                 bin_dir=None):
        bin_dir = bin_dir if bin_dir is not None else os.path.dirname(__file__)
        super().__init__(compressor_path=os.path.join(bin_dir, "fpack"),
                         decompressor_path=os.path.join(bin_dir, "funpack"),
                         param_dict=dict())

    @property
    def name(self):
        """Don't include the binary signature
        """
        return f"{self.__class__.__name__}{'__' if self.param_dict else ''}" \
               f"{'_'.join(f'{k}={v}' for k, v in self.param_dict.items())}"

    @property
    def label(self):
        return "FPACK - RICE"

    def get_compression_params(self, original_path, compressed_path, original_file_info):
        self.assert_valid_data_type(original_file_info=original_file_info)
        return f" -r  -O {compressed_path} {original_path}"

    def get_decompression_params(self, compressed_path, reconstructed_path, original_file_info):
        return f" -O {reconstructed_path} {compressed_path} "


class FPACK_HCOMPRESS(FPACK_Abstract):
    """fpack wrapper with hcompress compression.
    """
    def __init__(self,
                 bin_dir=None):
        bin_dir = bin_dir if bin_dir is not None else os.path.dirname(__file__)
        super().__init__(compressor_path=os.path.join(bin_dir, "fpack"),
                         decompressor_path=os.path.join(bin_dir, "funpack"),
                         param_dict=dict())

    @property
    def name(self):
        """Don't include the binary signature
        """
        return f"{self.__class__.__name__}{'__' if self.param_dict else ''}" \
               f"{'_'.join(f'{k}={v}' for k, v in self.param_dict.items())}"

    @property
    def label(self):
        return "FPACK - HCOMPRESS"

    def get_compression_params(self, original_path, compressed_path, original_file_info):
        self.assert_valid_data_type(original_file_info=original_file_info)
        return f" -h  -O {compressed_path} {original_path}"

    def get_decompression_params(self, compressed_path, reconstructed_path, original_file_info):
        return f" -O {reconstructed_path} {compressed_path} "



#!/usr/bin/env python3
"""PNG manipulation (e.g., curation) tools.
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2023/02/12"

import os
import subprocess
import tempfile

import imageio
import numpngw
import numpy as np
import pdf2image

import enb
from enb.sets import FileVersionTable
from enb.compression.wrapper import WrapperCodec
from enb.isets import load_array_bsq


class PNGWrapperCodec(WrapperCodec):
    """Raw images are coded into PNG before compression with the wrapper,
    and PNG is decoded to raw after decompression.
    """

    # pylint: disable=abstract-method

    def compress(self, original_path: str, compressed_path: str, original_file_info=None):
        img = load_array_bsq(
            file_or_path=original_path, image_properties_row=original_file_info)

        with tempfile.NamedTemporaryFile(suffix=".png") as tmp_file:
            numpngw.write_png(tmp_file.name, img)
            compression_results = super().compress(
                original_path=tmp_file.name,
                compressed_path=compressed_path,
                original_file_info=original_file_info)
            crs = self.compression_results_from_paths(
                original_path=original_path, compressed_path=compressed_path)
            crs.compression_time_seconds = max(
                0, compression_results.compression_time_seconds)
            crs.maximum_memory_kb = compression_results.maximum_memory_kb
            return crs

    def decompress(self, compressed_path, reconstructed_path,
                   original_file_info=None):
        with tempfile.NamedTemporaryFile(suffix=".png") as tmp_file:
            decompression_results = super().decompress(
                compressed_path=compressed_path,
                reconstructed_path=tmp_file.name)

            invocation = f"file {tmp_file.name}"
            status, output = subprocess.getstatusoutput(invocation)
            if status != 0:
                raise Exception(f"Status = {status} != 0.\n"
                                f"Input=[{invocation}].\n"
                                f"Output=[{output}]")
            img = imageio.imread(tmp_file.name, "png")
            img.swapaxes(0, 1)
            assert len(img.shape) in [2, 3, 4]
            if len(img.shape) == 2:
                img = np.expand_dims(img, axis=2)

            dtype = ">"
            dtype += "i" if original_file_info["signed"] else "u"
            dtype += f"{original_file_info['bytes_per_sample']}"

            enb.isets.dump_array_bsq(img, file_or_path=reconstructed_path,
                                     dtype=dtype)

            drs = self.decompression_results_from_paths(
                compressed_path=compressed_path,
                reconstructed_path=reconstructed_path)
            drs.decompression_time_seconds = decompression_results.decompression_time_seconds
            drs.maximum_memory_kb = decompression_results.maximum_memory_kb


class PNGCurationTable(FileVersionTable):
    """Given a directory tree containing PNG images, copy those images into
    a new directory tree in raw BSQ format adding geometry information tags to
    the output names recognized by `load_array_bsq`.
    """
    dataset_files_extension = "png"

    def __init__(self, original_base_dir, version_base_dir,
                 csv_support_path=None):
        """
        :param original_base_dir: path to the original directory
          (it must contain all indices requested later with self.get_df()).
          If None, options.base_datset_dir is used

        :param version_base_dir: path to the versioned base directory
          (versioned directories preserve names and structure within
          the base dir)

        :param csv_support_path: path to the file where results
          (of the versioned data) are to be
          long-term stored.
          If None, one is assigned by default based on options.persistence_dir.
        """
        super().__init__(version_base_dir=version_base_dir,
                         version_name=self.__class__.__name__,
                         original_base_dir=original_base_dir,
                         check_generated_files=False,
                         csv_support_path=csv_support_path)

    def version(self, input_path, output_path, row):
        """Transform PNG files into raw images with name tags
        recognized by isets.
        """
        png_to_raw(input_path, output_path)


def render_array_png(img, png_path):
    """Render an uint8 or uint16 image with 1, 3 or 4 components.
    :param img: image array indexed by [x,y,z].
    :param png_path: path where the png file is to be stored.
    """
    max_value = np.max(img)
    if img.dtype == np.uint8:
        pass
    elif any(img.dtype == t for t in (np.uint16, np.uint32, np.uint64)):
        if max_value <= 255:
            img = img.astype(np.uint8)
        elif max_value <= 65535:
            img = img.astype(np.uint16)
        else:
            raise ValueError(
                f"Invalid maximum value {max_value} for type {img.dtype}. "
                f"Not valid for PNG")
    else:
        raise ValueError(
            f"Image type {img.dtype} not supported for rendering into PNG. "
            f"Try np.uint8 or np.uint16.")

    if img.shape[2] not in {1, 3, 4}:
        raise ValueError(
            f"Number of components not valid. Image shape (x,y,z) = {img.shape}")
    if os.path.dirname(png_path):
        os.makedirs(os.path.dirname(png_path), exist_ok=True)
    imageio.imwrite(png_path, img.swapaxes(0, 1), format="png")


def raw_path_to_png(raw_path, png_path, image_properties_row=None):
    """Render an uint8 or uint16 raw image with 1, 3 or 4 components.

    :param raw_path: path to the image in raw format to render in png.
    :param png_path: path where the png file is to be stored.

    :param image_properties_row: if row_path does not contain geometry
      information, this parameter should be a dict-like object that indicates
      width, height, number of components, bytes per sample, signedness and
      endianness if applicable.
    """
    img = load_array_bsq(file_or_path=raw_path,
                                   image_properties_row=image_properties_row)
    render_array_png(img=img, png_path=png_path)

def png_to_raw(input_path, output_path) -> [str | os.PathLike]:
    """Read a png image from `input_path` 
    and write it to `<output_path without '.raw'>-<data_type>-ZxYxX.raw`,
    where `data_type` can be u8be or u16be, depending on the image contents.
    
    :return: the modified version of `output_path` actually used for the output.
    """
    with enb.logger.info_context(f"Versioning {input_path}"):
        img = imageio.v2.imread(input_path)
        if len(img.shape) == 2:
            img = img[:, :, np.newaxis]
        assert len(img.shape) == 3, \
            f"Invalid shape in read image {input_path}: {img.shape}"
        img = img.swapaxes(0, 1)
        if img.dtype == np.uint8:
            type_str = "u8be"
        elif img.dtype == np.uint16:
            type_str = "u16be"
        else:
            raise f"Invalid data type found in read image " \
                  f"{input_path}: {img.dtype}"
        output_path = f"{output_path[:-len(".raw")]}-{type_str}" \
                      f"-{img.shape[2]}x{img.shape[1]}x{img.shape[0]}.raw"
        enb.isets.dump_array_bsq(array=img, file_or_path=output_path)
        
        return output_path


class PDFToPNG(FileVersionTable):
    """Take all .pdf files in input dir and save them as .png files into
    output_dir, maintining the relative folder structure.
    """
    dataset_files_extension = "pdf"

    def __init__(self, input_pdf_dir, output_png_dir, csv_support_path=None):
        super().__init__(version_name="pdf_to_png",
                         original_base_dir=input_pdf_dir,
                         version_base_dir=output_png_dir,
                         csv_support_path=csv_support_path,
                         check_generated_files=True)

    def version(self, input_path, output_path, row):
        with enb.logger.info_context(
                f"{self.__class__.__name__}: {input_path} -> {output_path}...\n"):
            os.makedirs(os.path.dirname(output_path), exist_ok=True)
            imgs = pdf2image.convert_from_path(pdf_path=input_path)
            assert len(imgs) == 1
            imgs[0].save(output_path)


def pdf_to_png(input_dir, output_dir):
    """Take all .pdf files in input dir and save them as .png files into
    output_dir, maintining the relative folder structure.

    It is perfectly valid for input_dir and output_dir to point to the same
    location, but input_dir must exist beforehand.
    """
    with tempfile.NamedTemporaryFile() as tmp_file:
        PDFToPNG(input_pdf_dir=input_dir, output_png_dir=output_dir,
                 csv_support_path=tmp_file.name).get_df()



import textwrap
import enb.plugins


class MarlinPlugin(enb.plugins.PluginMake):
    name = "marlin"
    label = "Marlin-based image compressor"
    tags = {"data compression", "codec"}
    contrib_authors = ["Manuel Martínez Torres", "Miguel Hernández-Cabronero"]
    contrib_reference_urls = ["https://github.com/miguelinux314/marlin"]
    contrib_download_url_name = [
        ("https://github.com/miguelinux314/experiment-notebook/blob/dev/contrib/marlin_ubuntu20_git.zip?raw=true",
         "marlin_ubuntu20_git.zip")]
    extra_requirements_message = textwrap.dedent("""
            The cmake tool is needed to build this plugin, if not present already. 
            You can install cmake as follows: 

            * Ubuntu: `sudo apt install cmake`
            * MacOS:
                1. Install homebrew from http://brew.sh
                2. `brew install gcc@9`
            """)
    tested_on = {"linux"}



#!/usr/bin/env python3
"""Wrapper for the POT transform
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2020/05/25"

import os

from enb.config import options
from . import abstract_mhdc_transform

pot_transform_number = 2


class POTVersionTable(abstract_mhdc_transform.MHDCTransformTable):
    """Forward POT version table
    """

    @property
    def transform_number(self):
        return pot_transform_number


class InversePOTVersionTable(abstract_mhdc_transform.InverseMHDCTransformTable):
    """Inverse POT version table
    """

    @property
    def transform_number(self):
        return pot_transform_number


class POTLosslessCompressionExperiment(abstract_mhdc_transform.MDHCLosslessCompressionExperiment):
    @property
    def transform_number(self):
        return pot_transform_number


def apply_pot(input_dir, output_dir, forward_properties_csv=None, inverse_properties_csv=None,
              repetitions=10, run_sequential=True):
    """Apply the POT to input_dir and save the results to output_dir.

    :param input_dir: input directory to be transformed (recursively)
    :param output_dir: path to the output transformed dir
    :param forward_properties_csv: if not None, properties of the transformed images
      are stored here (including fwd transform time)
    :param inverse_properties_csv: if not None, properties of the reconstructed images
      are stored here (including inverse transform time)
    :param repetitions: number of repetitions used to calculate execution time
      (for both forward and inverse transform)
    :param run_sequential: if True, transformations are run in sequential mode
      (as opposed to parallel) so that time measurements can be taken

    :raises AssertionError: if transformation/reconstruction is not lossless

    :return: fwd_pot_df, inv_pot_df, i.e., the dataframes obtained with get_df
      for POTVersionTAble and InversePOTVersionTable, respectively.
    """
    return abstract_mhdc_transform.apply_transform(
        input_dir=input_dir, output_dir=output_dir,
        forward_class=POTVersionTable,
        inverse_class=InversePOTVersionTable,
        forward_properties_csv=forward_properties_csv,
        inverse_properties_csv=inverse_properties_csv,
        repetitions=repetitions, run_sequential=run_sequential)


if __name__ == '__main__':
    default_input_dir = "./green_book_corpus"
    default_output_dir = "/data/research-materials/pot_green_book"
    persistence_dir = "./persistence"
    repetition_count = 10
    run_sequential = True

    apply_pot(
        input_dir=os.path.abspath(os.path.realpath(
            options.base_dataset_dir if options.base_dataset_dir
            else default_input_dir)),
        output_dir=os.path.abspath(os.path.realpath(
            options.base_version_dataset_dir if options.base_version_dataset_dir
            else default_output_dir)),
        forward_properties_csv=os.path.join(persistence_dir, "pot_versioned_properties.csv"),
        inverse_properties_csv=os.path.join(persistence_dir, "ipot_versioned_properties.csv"),
        run_sequential=run_sequential,
        repetitions=repetition_count)



import os
import stat
import subprocess
import zipfile
import shutil

import enb.plugins


class TRPXPlugin(enb.plugins.Plugin):
    target_ref = "master"
    
    name = "trpx"
    label = "trpx (Terse/Prolix) codec"
    tags = {"data compression", "codec"}
    contrib_authors = ["Senik Matinyan"]
    contrib_reference_urls = ["https://github.com/senikm/trpx"]
    contrib_download_url_name = [
        (f"https://github.com/miguelinux314/trpx/archive/refs/heads/{target_ref}.zip",
         "trpx_codec_src.zip")]
    tested_on = {"linux"}

    @classmethod
    def install(cls, installation_dir, overwrite_destination=False):
        # Check that required tools are present before installing
        for binary in ("make", "cmake"):
            if shutil.which(binary) is None:
                raise ValueError(f"{binary} is not installed but it is need to install {cls.name} ({cls.label})")
            
        # Peform installation
        super().install(installation_dir=installation_dir, overwrite_destination=overwrite_destination)

    @classmethod
    def build(cls, installation_dir):
        # Copy files
        super().build(installation_dir=installation_dir)

        # Decompress the codec source
        assert len(cls.contrib_download_url_name) == 1
        with zipfile.ZipFile(os.path.join(installation_dir, cls.contrib_download_url_name[0][1])) as zfile:
            zfile.extractall(installation_dir)

        # Generate binaries (requires cmake and make)
        cwd = os.getcwd()
        try:
            src_dir = os.path.join(installation_dir, f"trpx-{cls.target_ref}")
            os.chdir(src_dir)

            # Generate build files
            invocation = f"cmake ."
            status, output = subprocess.getstatusoutput(invocation)
            if status != 0:
                raise Exception(f"Status = {status} != 0.\nInput=[{invocation}].\nOutput=[{output}]")

            # Compile
            invocation = "cmake --build ."
            status, output = subprocess.getstatusoutput(invocation)
            if status != 0:
                raise Exception(f"Status = {status} != 0.\nInput=[{invocation}].\nOutput=[{output}]")
            
            # Copy binary and make executable
            binary_path = os.path.join(installation_dir, "raw_codec") 
            shutil.copyfile(os.path.join(src_dir, "src", "raw_codec"), binary_path)
            os.chmod(binary_path, os.stat(binary_path).st_mode | stat.S_IEXEC)
            
        finally:
            os.chdir(cwd)



#!/usr/bin/env python3
"""Wrapper for Michael Dipperstein's RLE codec, originally downloaded from
https://github.com/MichaelDipperstein/rle. Also see https://michaeldipperstein.github.io/rle.html.
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2024/01/15"

import os
from enb import icompression


class RLECodec(icompression.WrapperCodec, icompression.LosslessCodec):
    """Wrapper for Michael Dipperstein's RLE codec, originally downloaded from
    https://michaeldipperstein.github.io/index.html. Also see https://michaeldipperstein.github.io/rle.html.
    """

    def __init__(self, packbits=False, compressor_path=None, decompressor_path=None):
        """
        :param packbits: if True, the packbits variant is employed
        """
        if compressor_path is None:
            compressor_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "rle")
        if decompressor_path is None:
            decompressor_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "rle")
        super().__init__(compressor_path=compressor_path, decompressor_path=decompressor_path,
                         param_dict=dict(packbits=packbits))

    def get_compression_params(self, original_path, compressed_path, original_file_info):
        return f"-c {'-v' if self.param_dict['packbits'] else ''} -i {original_path} -o {compressed_path}"

    def get_decompression_params(self, compressed_path, reconstructed_path, original_file_info):
        return f"-d {'-v' if self.param_dict['packbits'] else ''} -i {compressed_path} -o {reconstructed_path}"

    @property
    def label(self):
        return f"RLE{' packbits' if self.param_dict['packbits'] else ''}"
